/*
***************************************************************************
**                                                                       **
**	Restart manager  -  Extension for Pale Moon                          **
**                                                                       **
**	Main code for main window and options dialog                         **
**                                                                       **
**	Copyright(C) 2018 by George Malamas                                  **
**                                                                       **
**	Free software for all people                                         **
**                                                                       **
**	Last edit: July 1, 2018                                              **
**                                                                       **
**	Release number: 3                                                    **
**                                                                       **
***************************************************************************
*/
// To do: Make the code compatible with strict mode.
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Functions_and_function_scope/Strict_mode
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Functions_and_function_scope/Strict_mode/Transitioning_to_strict_mode
// Any help or advises will be very appreciated :-)
// "use strict";

var cc = Components.classes;    // Workaround for "redeclaration of const Cc,Ci,Cu" error in some
var ci = Components.interfaces; // app versions, while some other versions require these
var cu = Components.utils;		// declarations, even for the overlay part of our code..

// this.EXPORTED_SYMBOLS = ["_RestMan"]; // (In combination with importing this file in our js component)
// It works fine, but causes almost quadruple of memory commitment..

var _RestMan = {
	/*<Shared stuff>**********************************/
	// The services we need:
	appInfo:         cc["@mozilla.org/xre/app-info;1"].getService(ci.nsIXULAppInfo),
	appStartup:      cc["@mozilla.org/toolkit/app-startup;1"].getService(ci.nsIAppStartup),
	dirSrvc:         cc["@mozilla.org/file/directory_service;1"].getService(ci.nsIProperties),
	obsSrvc:         cc["@mozilla.org/observer-service;1"].getService(ci.nsIObserverService),
	prefSrvc:        cc["@mozilla.org/preferences-service;1"].getService(ci.nsIPrefService),
	promptSrvc:      cc["@mozilla.org/embedcomp/prompt-service;1"].getService(ci.nsIPromptService),
	restmanSrvc:     null, // First, we check below if appVer version is >=3. Otherwise, we'll be broken in appVer 1.5 & 2..
	strings:         cc["@mozilla.org/intl/stringbundle;1"].getService(ci.nsIStringBundleService),
	timer:           cc["@mozilla.org/timer;1"].createInstance(ci.nsITimer),
	winMed:          cc["@mozilla.org/appshell/window-mediator;1"].getService(ci.nsIWindowMediator),
	// App info:
	appName:         "Pale Moon",
	appVer:          null,
	// Our prefBranch:
	rmPrefBranch:    cc["@mozilla.org/preferences-service;1"].getService(ci.nsIPrefService)
						.getBranch("extensions.restartmanager."),
	// Browser's prefBranch:
	brPrefBranch:    cc["@mozilla.org/preferences-service;1"].getService(ci.nsIPrefService)
						.getBranch("browser."), // Pale Moon
	// Most of our prefs:
	menuItemIcon:    true,
	iconStyle:       0,
	rmButtonStyle:   1, // appVer 4+.
	rmSmallButton:   false, // Use small button allways, appVer 4+.
	askOnRestart:    true,
	sessionRestore:  false,
	warnOnCloseMT:   false, // multiple tabs.
	warnOnCloseMPT:  true, // multiple private tabs, appVer 20+ with "Private Tab" extension.
	detectOpenDlgs:  true, // Detection of opened dialogs on restart, exit or close of the last browser window.
	restoreUserSet:  false, // User settings we change temporarily on a specific case.
	restWarnOnQuit:  false, // Restore "browser.warnOnQuit" pref.
	codeMonitoring:  false, // Can be activated in "about:config"("alertsEnabled"), to enable code flow monitoring in selected points, implemented by a sequence of alert messages.
	codMonChkBoxHid: true, // Checkbox in options dialog(hidden by default, "codeMonCheckbox_hidden"), which controls the above pref ("alertsEnabled").
	restShortcutVis: true, // Restart shortcut indication in restart menu items.
	restartShortcut: true, // Restart shortcut active.
	shortcutKeyMod:  "alt",
	shortcutChar:    "R",
	invalCaches:     false, // Invalidate Caches On Restart
	invCachHghlght:  false, // Highlight Invalidate Caches
	// App version specific strings & string sources:
	confirmLabel:    "Confirm", // All app versions.
	neverAskLabel:   "Do not ask next time",
	restartLabel:    "Restart", // All app versions.
	restartAccKey:   "R",       //
	restartTitle:    "Restart Pale Moon", //appVer >=4.
	restartMessage:  "Restart Firefox now?", // appVer 1.5 & 2. Old Pale Moon versions are not fully "cleared" of firefox.
	optionsTitle:    "Restart manager - Options", // All app versions.
	optionsLabel:    "Options", // All app versions.
	optionsAccessk:  "O",       //
	appearanceLabel: "Appearance", //appVer >=6.
	// Count of windows & tabs, on non-private and private mode:
	allBrWinCount:   0,
	nonPrvWinCount:  0,
	privWinCount:    0,
	nPrWinVPrTbCnt:  0, // Non-private windows with all their visible tabs private. Compatibility with "Private Tab" extension.
	nPrWnNprVTbCnt:  0, // Non-private windows with at least 1 non-private visible tab. Compatibility with "Private Tab" extension.
	allTabCount:     0,
	visTabCount:     0, // Visible tabs.
	nonPrVisTabCnt:  0, // Non-private visible tabs.
	prvVisTabCount:  0, // Private visible tabs. Compatibility with "Private Tab" extension.
	allTabCountCW:   0, // All tabs in current window.
	nPrVisTabCntCW:  0, // Non-private visible tabs in current window.
	prvVisTabCntCW:  0, // Private visible tabs in current window. Compatibility with "Private Tab" extension.
	// Note: Visible tabs = tabs in the current tab group.
	allWinPrivate:   true,
	isWinPrivate:    true, // Current window.
	allTabsPrivate:  false, // Compatibility with "Private Tab" extension.
	allTabsPrvCW:    false, // Current window, compatibility with "Private Tab" extension.
	privateTabExt:   false, // Presence or status of "Private Tab" extension.
	// Helper objects
	openedDialog:    null, // Detection of opened dialogs.
	dnsPrefObserver: false, // "Do restart no restore" pref observer.
	setPrefObserver: false, // Apply ui settings pref observer.
	/*********************************</Shared stuff>*/

	// Our main localized strings are included, or have been composed from included strings in each app version.
	// Due to this localization method, some of our strings may not be so appropriate, or may be poor,
	// or there may be problems in some languages, especially in eastern languages.
	// For this reason also, due to lack of strings, we use some images instead of labels in our options dialog.

	onLoadStrings: function() {
	// Definition of strings/variables needed on app loading.
		var brandBundle = _RestMan.strings.createBundle("chrome://branding/locale/brand.properties");
		var appName = brandBundle.GetStringFromName("brandShortName");
		if(appName !== null || appName !== undefined) {
			_RestMan.appName = _RestMan.appInfo.name;
		};
		_RestMan.appVer = parseFloat(_RestMan.appInfo.version); // parseInt
		if(_RestMan.appVer >=3) {
			_RestMan.restmanSrvc = cc["@restart_manager/chania.gr;1"].getService().wrappedJSObject; // Our js component.
		};
		_RestMan.codeMonitoring = _RestMan.rmPrefBranch.getBoolPref("alertsEnabled");
		_RestMan.menuItemIcon = _RestMan.rmPrefBranch.getBoolPref("menuItemIcon");
		_RestMan.iconStyle = _RestMan.rmPrefBranch.getIntPref("iconStyle");
		_RestMan.rmButtonStyle = _RestMan.rmPrefBranch.getIntPref("buttonStyle");
		_RestMan.rmSmallButton = _RestMan.rmPrefBranch.getBoolPref("allways_smallBtn"); // Used in appVer 4+.
		// We get from here the label and access key for our button's context menu.
		var appOptionsContext = _RestMan.winMed.getMostRecentWindow("navigator:browser")
			.document.getElementById("menu_preferences"); // It is accessed also by our js component.
			if(appOptionsContext) { // Menu bar "Tools --> Options or Preferences".
				_RestMan.optionsLabel = appOptionsContext.getAttribute("label");
				_RestMan.optionsAccessk = appOptionsContext.getAttribute("accesskey");
			}; //
		_RestMan.detectOpenDlgs = _RestMan.rmPrefBranch.getBoolPref("openDlg_detection");
		_RestMan.restShortcutVis = _RestMan.rmPrefBranch.getBoolPref("restart_shortcutVis");
		_RestMan.restartShortcut = _RestMan.rmPrefBranch.getBoolPref("restart_shortcut");
		if(_RestMan.rmPrefBranch.getCharPref("shortcutKey_modifier") == "alt-option") {
			_RestMan.shortcutKeyMod = "alt";
		} else if(_RestMan.rmPrefBranch.getCharPref("shortcutKey_modifier") == "shift") {
			_RestMan.shortcutKeyMod = "shift";
		};
		_RestMan.invCachHghlght = _RestMan.rmPrefBranch.getBoolPref("invalCaches_highlight");
		_RestMan.shortcutChar = _RestMan.rmPrefBranch.getComplexValue("shortcut_char", ci.nsISupportsString).data;
		if(_RestMan.appVer >=20) {
			var browserBundle = _RestMan.strings.createBundle("chrome://browser/locale/browser.properties");
			_RestMan.restartLabel = browserBundle.GetStringFromName("safeModeRestartButton"); // Restart
			var prefsBundle = _RestMan.strings.createBundle("chrome://browser/locale/preferences/preferences.properties");
			_RestMan.restartTitle = prefsBundle.formatStringFromName("shouldRestartTitle", [_RestMan.appName], 1); // Restart brandShortName
			_RestMan.restoreUserSet = _RestMan.rmPrefBranch.getBoolPref("restore_UserSettings");
			_RestMan.restWarnOnQuit = _RestMan.rmPrefBranch.getBoolPref("restore_warnOnQuitPref");
		} else if(_RestMan.appVer <20 && _RestMan.appVer >=3) {
			var quitBundle = _RestMan.strings.createBundle("chrome://browser/locale/quitDialog.properties");
			_RestMan.restartLabel = quitBundle.GetStringFromName("restartTitle").replace(/&/,"").replace(/\(.\)/,""); // Restart
			if(_RestMan.appVer >=4) {
				_RestMan.restartTitle = quitBundle.formatStringFromName("restartDialogTitle", [_RestMan.appName], 1); // Restart brandShortName
			};
		} else if(_RestMan.appVer <3 && _RestMan.appVer >1.5) {
			var extBundle = _RestMan.strings.createBundle("chrome://mozapps/locale/extensions/extensions.properties");
			var capital = extBundle.GetStringFromName("restartButton").replace(/ +%S|%S +/,"").charAt(0).toLocaleUpperCase(); // R
			_RestMan.restartLabel = extBundle.GetStringFromName("restartButton").replace(/ +%S|%S +/,"").replace(/^./,capital); // Restart
		} else if(_RestMan.appVer <2) {
			// The most suitable source to get the "Restart" string is the profileSelection.properties file. However, in this file the property
			// "restartTitle=" has a wrong string in many locales, including en-US("restartTitle=Close %S" instead of "restartTitle=Restart %S").
			var updateBundle = _RestMan.strings.createBundle("chrome://mozapps/locale/update/updates.properties");
			var capital = updateBundle.GetStringFromName("restartButton").replace(/ +%S|%S +/,"").charAt(0).toLocaleUpperCase(); // R
			_RestMan.restartLabel = updateBundle.GetStringFromName("restartButton").replace(/ +%S|%S +/,"").replace(/^./,capital); // Restart Now
		};
		var exitLabel = _RestMan.winMed.getMostRecentWindow("navigator:browser").document.getElementById("menu_FileQuitItem")
			.getAttribute("label"); // "File" menu --> "Exit"
		var exitAccKey = _RestMan.winMed.getMostRecentWindow("navigator:browser").document.getElementById("menu_FileQuitItem")
			.getAttribute("accesskey");
		var ak = new RegExp(exitAccKey, "i");
		if(ak.test(exitLabel)) {
			_RestMan.restartAccKey = _RestMan.restartLabel.match(/^./);
		} else { // Probably it's an eastern locale (or it is a localization mistake!).
			_RestMan.restartAccKey = "R";
		};
	},

	_RestManOnLoad: function RestMan_RestManOnLoad() {
		_RestMan.onLoadStrings();
		if(_RestMan.appVer <20 && _RestMan.appVer >2 && window.document.getElementById("commonDialog")) {
			if(_RestMan.restmanSrvc.getSetVal("restartFlag", "getV")) { // get value.
			// We patch "on the fly" the message of restart dialog in appVer 3-19, to avoid
			// repeating here the code of the dialog, just to modify slightly its message..
			// Details for this can be found in "restartDlg3to19()" function.
				var browserBundle = _RestMan.strings.createBundle("chrome://browser/locale/browser.properties");
				var msgAddition;
				if(_RestMan.appVer >=4) {
					msgAddition = browserBundle.GetStringFromName("addonInstallRestartButton"); // Restart Now
				} else if(_RestMan.appVer <4) {
					var updatesBundle = _RestMan.strings.createBundle("chrome://mozapps/locale/update/updates.properties");
					if(_RestMan.appVer <4 && _RestMan.appVer >3) {
						msgAddition = updatesBundle.formatStringFromName("restartNowButton", [_RestMan.appName], 1); // Restart brandShortName
					} else if(_RestMan.appVer ==3) {
						msgAddition = updatesBundle.formatStringFromName("restartButton", [_RestMan.appName], 1); // Restart brandShortName Now
					};
				};
				var questionMark = browserBundle.GetStringFromName("droponhomemsg").match(/.$/); // ?
				var dlgMessage = window.document.getElementById("info.body");
				var textToAdd = window.document.createTextNode("\n" + msgAddition + questionMark); // Restart (brandShortName) (Now)?
				if(dlgMessage) {
					dlgMessage.appendChild(textToAdd);
				};
				_RestMan.restmanSrvc.getSetVal("restartFlag", "setF"); // set false.
				return;
			};
		};
		function addMenuPopUpLis(aTarget) {
			var targetMenu;
			if(aTarget == "appMenu") {
				targetMenu = window.document.getElementById("appmenu-popup");
			} else if(aTarget == "fileMenu") {
				targetMenu = window.document.getElementById("menu_FilePopup");
			};
			if(targetMenu) {
				targetMenu.addEventListener("popupshowing", function onMenuPopUpShowing() {
					_RestMan.sharedStrings();
					_RestMan.winTabCounter();
					var targetMenuItem;
					if(targetMenu.id == "appmenu-popup") { // "targetMenuItem" must be defined here. In case of remove/disable/change of
					// our shortcut, "targetMenuItem" is destroyed/reconstructed. Otherwise "onMenuPopUpShowing()" will lose its target.
						targetMenuItem = window.document.getElementById("restman-appMenuItem");
					} else if(targetMenu.id == "menu_FilePopup") {
						targetMenuItem = window.document.getElementById("restman-fileMenuItem");
					};
					if(_RestMan.appVer >19) {
						if(_RestMan.nonPrvWinCount >=1) {
							if((_RestMan.askOnRestart || (!_RestMan.askOnRestart && _RestMan.warnOnCloseMT)) &&
								_RestMan.nonPrVisTabCnt >1) { // The action will show a dialog.
									targetMenuItem.label = _RestMan.restartLabel + "…";
							} else if(_RestMan.nonPrVisTabCnt <2 || (!_RestMan.askOnRestart && !_RestMan.warnOnCloseMT) ||
								(!_RestMan.askOnRestart && _RestMan.sessionRestore)) { // The action will cause an immediate restart.
								targetMenuItem.label = _RestMan.restartLabel;
							};
						} else {
							if(_RestMan.warnOnCloseMPT && _RestMan.allTabCount >1) { // The action will show a dialog.
								targetMenuItem.label = _RestMan.restartLabel + "…";
							} else if(_RestMan.allTabCount <2 || !_RestMan.warnOnCloseMPT) { // The action will cause an immediate restart.
								targetMenuItem.label = _RestMan.restartLabel;
							};
						};
					} else if(_RestMan.appVer <20 && _RestMan.appVer >3) {
						if(_RestMan.askOnRestart && _RestMan.allTabCount >1 && 
							!cc["@mozilla.org/privatebrowsing;1"].getService(ci.nsIPrivateBrowsingService)
							.privateBrowsingEnabled) { // The action will show a dialog.
							targetMenuItem.label = _RestMan.restartLabel + "…";
						} else if(!_RestMan.askOnRestart || _RestMan.allTabCount <2 ||
							cc["@mozilla.org/privatebrowsing;1"].getService(ci.nsIPrivateBrowsingService)
							.privateBrowsingEnabled) { // The action will cause an immediate restart.
							targetMenuItem.label = _RestMan.restartLabel;
						};
					} else if(_RestMan.appVer <3.5) {
						if(_RestMan.askOnRestart && _RestMan.allTabCount >1) { // The action will show a dialog.
							targetMenuItem.label = _RestMan.restartLabel + "…";
						} else if(!_RestMan.askOnRestart || _RestMan.allTabCount <2) { // The action will cause an immediate restart.
							targetMenuItem.label = _RestMan.restartLabel;
						};
					};
				}, false);
			};
		};
		var appMenuItem = window.document.getElementById("restman-appMenuItem");
		if(appMenuItem) {
			appMenuItem.label = _RestMan.restartLabel;
			appMenuItem.setAttribute("accesskey", _RestMan.restartAccKey);
			if(_RestMan.menuItemIcon) {
				appMenuItem.setAttribute("menuicon", true);
			};
			appMenuItem.setAttribute("iconStyle", _RestMan.iconStyle);
			if(_RestMan.invCachHghlght) {
				appMenuItem.setAttribute("highlightMenuItem", true);
			};
			if(_RestMan.appVer >=22) {
				appMenuItem.setAttribute("appVer22-28", "");
			};
			addMenuPopUpLis("appMenu");
		};
		var fileMenuItem = window.document.getElementById("restman-fileMenuItem");
		if(fileMenuItem) {
			fileMenuItem.label = _RestMan.restartLabel;
			fileMenuItem.setAttribute("accesskey", _RestMan.restartAccKey);
			if(_RestMan.menuItemIcon) {
				fileMenuItem.setAttribute("menuicon", true);
			};
			fileMenuItem.setAttribute("iconStyle", _RestMan.iconStyle);
			if(_RestMan.invCachHghlght) {
				fileMenuItem.setAttribute("highlightMenuItem", true);
			};
			if(_RestMan.appVer <3) {
				fileMenuItem.setAttribute("appVer1v5-2", "");
			};
			if(_RestMan.appVer >=22) {
				fileMenuItem.setAttribute("appVer22-28", "");
			};
			addMenuPopUpLis("fileMenu");
		};
		var restmanKeyset = window.document.getElementById("restmanKeyset");
		var restartShortcutElem = window.document.getElementById("key_restmanKeyCombo");
		if(restmanKeyset && !_RestMan.restartShortcut) {
			restmanKeyset.parentNode.removeChild(restmanKeyset);
			if(appMenuItem && appMenuItem.hasAttribute("key")) {
				appMenuItem.removeAttribute("key");
			};
			if(fileMenuItem && fileMenuItem.hasAttribute("key")) {
				fileMenuItem.removeAttribute("key");
			};
		} else if(restartShortcutElem && _RestMan.restartShortcut) { // Set shortcut attributes.
			restartShortcutElem.setAttribute("modifiers", "accel," + _RestMan.shortcutKeyMod);
			restartShortcutElem.setAttribute("key", _RestMan.shortcutChar);
			if(!_RestMan.restShortcutVis) {
				if(appMenuItem && appMenuItem.hasAttribute("key")) {
					appMenuItem.removeAttribute("key");
				};
				if(fileMenuItem && fileMenuItem.hasAttribute("key")) {
					fileMenuItem.removeAttribute("key");
				};
			};
		};
		if(window.document.getElementById("appmenu_restart")) { // Removing of Pale Moon's native restart menu items, introduced in v27.5.0.
			window.document.getElementById("appmenu_restart").parentNode.removeChild(window.document.getElementById("appmenu_restart")); //
		};																																 //
		if(window.document.getElementById("menu_restart")) {																			 //
			window.document.getElementById("menu_restart").parentNode.removeChild(window.document.getElementById("menu_restart"));		 //
		};																																 //
		var restartButton = window.document.getElementById("restman-button");
		if(restartButton) {
			restartButton.label = _RestMan.restartLabel;
			restartButton.setAttribute("tooltiptext", _RestMan.restartLabel);
			restartButton.setAttribute("iconStyle", _RestMan.iconStyle);
			if(_RestMan.invCachHghlght) {
				restartButton.setAttribute("highlightBtn", true);
			};
			if(_RestMan.appVer >=4) {
				if(_RestMan.rmButtonStyle == 0)  {
					restartButton.setAttribute("class", "toolbarbutton-1");
				};
				if(_RestMan.rmSmallButton)  {
					restartButton.setAttribute("smallButton", true);
				};
				if(_RestMan.appVer >=22) {
					restartButton.setAttribute("appVer22-28", "");
				};
				if(_RestMan.appVer >=14) {
					restartButton.setAttribute("appVer14-28", "");
				};
				if(_RestMan.appVer <14) {
					restartButton.setAttribute("appVer4-13", "");
				};
				// We set the width of our button's label to be always as when the label is bold, to avoid the button to be
				// expanded/reduced when the label becomes bold/normal. This way, the extra space of label when it is normal
				// "absorbs" the extra width when it is bold.
				var hostToolbar = window.document.getElementById(restartButton.parentNode.parentNode.id);
				function setLblWidth(aParam) {
					if(!_RestMan.invCachHghlght) {
						restartButton.setAttribute("highlightBtn", true); // Makes the label "bold"
					};
					var btnLabel = restartButton.ownerDocument.getAnonymousNodes(restartButton)[1];
					// var boldWidth = Math.round(parseInt(window.getComputedStyle(btnLabel).getPropertyValue("width")));
					var boldWidth = btnLabel.clientWidth + 1; // Simpler/accurate than above, in all supported appVer.
					if(btnLabel.getAttribute("class") == "toolbarbutton-text") { // Just for peace of mind..
						btnLabel.setAttribute("style", "width: "+boldWidth + "px;"+"");
					};
					if(!_RestMan.invCachHghlght) {
						restartButton.removeAttribute("highlightBtn");
					};
					if(aParam == "restTVS") { // Restore toolbar's visibility status.
						hostToolbar.setAttribute("collapsed", true);
					};
					if(aParam == "restVTVS0") { // Restore Vertical Toolbar's visibility status.
						_RestMan.prefSrvc.getBranch("extensions.verticaltoolbar.").setIntPref("display", 0);
						hostToolbar.parentNode.setAttribute("collapsed", true);
					};
					if(aParam == "restVTVS2-autohide") { // Restore Vertical Toolbar's visibility status.
						_RestMan.prefSrvc.getBranch("extensions.verticaltoolbar.").setIntPref("display", 2);
						hostToolbar.parentNode.setAttribute("autohide", true);
					};
					if(aParam == "restVTVS2-noAutohide") { // Restore Vertical Toolbar's visibility status.
						_RestMan.prefSrvc.getBranch("extensions.verticaltoolbar.").setIntPref("display", 2);
					};
				};
				var toolbarMode = hostToolbar.getAttribute("mode");
				if(toolbarMode == "full" || toolbarMode == "text") {
					if(hostToolbar.id !== "vertical-toolbar") {
						if(hostToolbar.hasAttribute("collapsed") && hostToolbar.getAttribute("collapsed") == "true") {
							hostToolbar.removeAttribute("collapsed");
							setLblWidth("restTVS");
						} else {
							setLblWidth();
						};
					} else if(hostToolbar.id == "vertical-toolbar") { // Vertical Toolbar.
					// A bit hassle for Vertical Toolbar, but this extension is one of my favorites :-)
						var vertTlbrPrefs = _RestMan.prefSrvc.getBranch("extensions.verticaltoolbar.");
						if(vertTlbrPrefs.getIntPref("display") == 1) { // Always show.
							setLblWidth();
						} else if(vertTlbrPrefs.getIntPref("display") == 0) { // Never show.
							vertTlbrPrefs.setIntPref("display", 1);
							hostToolbar.parentNode.removeAttribute("collapsed");
							setLblWidth("restVTVS0");
						} else if(vertTlbrPrefs.getIntPref("display") == 2) { // Automatically hide.
							vertTlbrPrefs.setIntPref("display", 1);
							if(hostToolbar.parentNode.hasAttribute("autohide")) { // Button "Always Show" is not checked.
								hostToolbar.parentNode.removeAttribute("autohide");
								setLblWidth("restVTVS2-autohide");
							} else if(!hostToolbar.parentNode.hasAttribute("autohide")) { // Button "Always Show" is checked.
								setLblWidth("restVTVS2-noAutohide");
							};
						};
					};
				};
				if(_RestMan.appInfo.QueryInterface(ci.nsIXULRuntime).OS =="Linux") {
					restartButton.setAttribute("Linux", "");
				};
			} else {
				restartButton.setAttribute("class", "toolbarbutton-1");
			};
		};
		var rmBtn_OptionsContext = window.document.getElementById("restman-launchOptionsItem");
		if(rmBtn_OptionsContext) {
			rmBtn_OptionsContext.label = _RestMan.optionsLabel;
			rmBtn_OptionsContext.setAttribute("accesskey", _RestMan.optionsAccessk);
		};
		if(_RestMan.appVer <20 && _RestMan.appVer >=3) {
			if(_RestMan.rmPrefBranch.getBoolPref("firstRun")) {
				// We do this here, because if we put "browser.warnOnRestart" in our defaults
				// file, some versions of pm don't respect user setting, reverting this pref
				// to our default value in every startup.
				_RestMan.brPrefBranch.setBoolPref("warnOnRestart", true);
				_RestMan.rmPrefBranch.setBoolPref("firstRun", false);
			};
		} else {
			_RestMan.rmPrefBranch.setBoolPref("firstRun", false); // Just for the sake of logic..
		};
		if(_RestMan.appVer >=20) {
			var browserWin = window.document.getElementById("main-window");
			// In a mixed session (normal & private), despite of pm defaults, we want to save the session of the
			// last closed non-private window, even on exiting or restarting pm, from within private browsing mode.
			// Details for this can be found in "sessionSaveHelper" function.
			// For this, we can't use the "unload" event, because, when it is dispatched is too late for the session
			// to be saved.
			// Also, we may have to perform some actions on closing of the last browser window. Details for this
			// can be found in "onQuitRequest" function.
			// We can't use listeners for "close" & "DOMWindowClose" events, because in case of warning (pm default)
			// or quit dialog, the listener captures the event after the dialog is closed, in both cases of dialog
			// accept/cancel.
			// There are also two important reasons we want to get the events before pm:
			// 1. Pm has only one warning dialog on attempt to close multiple tabs, which is used in both non-private
			// and private or mixed browsing modes. If this dialog is disabled, there is no any warning in all modes.
			// Also, this dialog is not shown on closing the last browser window, if it is private, or on pm exiting from 
			// private browsing mode.
			// We show four different warning dialogs: 1. - On restart in non-private or mixed browsing mode, 2. - on closing
			//   a private window, 3. - on closing the last browser window which is private or all of its tabs are private,
			//   4. - on exiting pm in private browsing mode. This is important in our opinion, because private windows get
			//   lost for good after closing them..
			// 2. Pm does not recognize the case where a window is non-private but all of its tabs are private
			// (using "Private tab" extension), showing its quit dialog (if enabled) which offers "Save and exit" option.
			// We handle properly this case, showing our warning dialog instead.
			// So, on closing a window, we check first for privacy and act accordingly: We let pm do its job as usual or
			// we take the control.
			// On pm exiting, this check is performed of the observer in our js component.
			// The following approach seems to solve our problems reliably, because it works and does not affect the normal
			// operation of pm.
			if(browserWin) {
				if(browserWin.getAttribute("windowtype") == "navigator:browser") { // We do this because some other
					// windows have also id="main-window" (scratchpad.xul, pageInfo.xul, filepicker.xul). Of course,
					// our overlay is applied only in browser.xul, but we feel a bit.. insecure.. :-)
					// When the window is closed with the close button:
					browserWin.setAttribute("onclose", "return _RestMan.onBrWinClose(event);");
					// When the window is closed with Ctrl+Shift+w:
					// (In case of warning dialog, "WindowIsClosing()" (browser.js) function returns true only if
					// the dialog is accepted)
					browserWin.parentNode.getElementById("cmd_closeWindow").setAttribute("oncommand",
						"if(_RestMan.winTabCounter(\"checkPrivacy\")) {\
							if(WindowIsClosing()) {\
								_RestMan.winTabCounter(\"brHandlesClosing\");\
								window.close();\
							};\
						} else {\
							_RestMan.winTabCounter(\"rmHandlesClosing\");\
						};"
					);
				};
			}; //
			// In case there are no browser windows and the user decides to open again
			// a browser window, we also have to register again our observer:
			window.setTimeout(function() {
				if(!_RestMan.restmanSrvc.getSetVal("observerIsRegistered", "getV")) {
					_RestMan.restmanSrvc.observerRegister();
				};
			}, 200); //
			if(browserWin) {
				if(browserWin.getAttribute("windowtype") == "navigator:browser") {
					window.document.getElementById("helpSafeMode").setAttribute("oncommand", "_RestMan.safeModeRestart();");
					if(window.document.getElementById("appmenu_safeMode")) {
						window.document.getElementById("appmenu_safeMode").setAttribute("oncommand", "_RestMan.safeModeRestart();");
					};
				};
			};
		};
		if(_RestMan.appVer >=4) {
			function disableDevPrefs() {
				if(_RestMan.rmPrefBranch.getBoolPref("invalidate_caches")) {
					_RestMan.rmPrefBranch.setBoolPref("invalidate_caches", false);
				};
				if(_RestMan.rmPrefBranch.getBoolPref("invalCaches_highlight")) {
					_RestMan.applyUISettings();
					_RestMan.rmPrefBranch.setBoolPref("invalCaches_highlight", false);
				};
				if(_RestMan.prefSrvc.getBranch("nglayout.").getBoolPref("debug.disable_xul_fastload")) {
					_RestMan.prefSrvc.getBranch("nglayout.").setBoolPref("debug.disable_xul_fastload", false);
				};
				if(_RestMan.prefSrvc.getBranch("nglayout.").getBoolPref("debug.disable_xul_cache")) {
					_RestMan.prefSrvc.getBranch("nglayout.").setBoolPref("debug.disable_xul_cache", false);
				};
			};
			var addonListeners = {
				onDisabling: function(addon) {
					if(addon.id == "restartmanager@chania.gr") {
						disableDevPrefs();
					};
				},
				onUninstalling: function(addon) {
					if(addon.id == "restartmanager@chania.gr") {
						disableDevPrefs();
					};
				}
			};
			if(!_RestMan.restmanSrvc.getSetVal("disableUninstallLis", "getV")) { // get value.
				// Ensuring only one set of listeners per session, regardless of opened windows.
				cu["import"]("resource://gre/modules/AddonManager.jsm", _RestMan);
				AddonManager.addAddonListener(addonListeners);
				_RestMan.restmanSrvc.getSetVal("disableUninstallLis", "setT"); // set true.
			};
		};
	},

	sharedStrings: function() {
	// Definition of some shared strings/variables, used in dialogs and restart process.
		_RestMan.onLoadStrings();
		var cdBundle = _RestMan.strings.createBundle("chrome://global/locale/commonDialogs.properties");
		var restmanBundle = _RestMan.strings.createBundle("chrome://restartmanager/locale/restartmanager.properties");
		var browserBundle = _RestMan.strings.createBundle("chrome://browser/locale/browser.properties");
		var restmanName = restmanBundle.GetStringFromName("extensions.restartmanager@chania.gr.name"); // Restart manager
		var optionsStr = browserBundle.GetStringFromName("popupWarningButton"); // Options
		_RestMan.optionsTitle = restmanName + " - " + optionsStr; // Restart manager - Options
		_RestMan.confirmLabel = cdBundle.GetStringFromName("ConfirmCheck"); // Confirm
		_RestMan.invalCaches = _RestMan.rmPrefBranch.getBoolPref("invalidate_caches");
		_RestMan.codMonChkBoxHid = _RestMan.rmPrefBranch.getBoolPref("codeMonCheckbox_hidden");
		if(_RestMan.appVer >=6) {
			var extensionsBundle = _RestMan.strings.createBundle("chrome://mozapps/locale/extensions/extensions.properties");
			_RestMan.appearanceLabel = extensionsBundle.GetStringFromName("type.theme.name"); // Appearance
		};
		if(_RestMan.appVer >=20) {
			_RestMan.askOnRestart = _RestMan.rmPrefBranch.getBoolPref("askOnRestart");
			_RestMan.sessionRestore = _RestMan.rmPrefBranch.getBoolPref("restoreSession_onRestart");
			_RestMan.warnOnCloseMT = _RestMan.rmPrefBranch.getBoolPref("warnOnClose_multipleTabs");
			_RestMan.warnOnCloseMPT = _RestMan.rmPrefBranch.getBoolPref("warnOnClose_multiplePrvTabs");
		};
		if(_RestMan.appVer >=23) {
			var quitBundle = _RestMan.strings.createBundle("chrome://browser/locale/quitDialog.properties");
			_RestMan.neverAskLabel = quitBundle.GetStringFromName("neverAsk2"); // &Do not ask next time
		} else if(_RestMan.appVer <23 && _RestMan.appVer >=20) {
			var quitBundle = _RestMan.strings.createBundle("chrome://browser/locale/quitDialog.properties");
			_RestMan.neverAskLabel = quitBundle.GetStringFromName("neverAsk"); // Do not ask next time
		} else if(_RestMan.appVer <20 && _RestMan.appVer >=3) {
			var quitBundle = _RestMan.strings.createBundle("chrome://browser/locale/quitDialog.properties");
			_RestMan.askOnRestart = _RestMan.brPrefBranch.getBoolPref("warnOnRestart");
			_RestMan.neverAskLabel = quitBundle.GetStringFromName("neverAsk"); // Do not ask next time
		} else if(_RestMan.appVer <3) {
			var askBundle = _RestMan.strings.createBundle("chrome://global/locale/nsProgressDialog.properties");
			var updateBundle = _RestMan.strings.createBundle("chrome://mozapps/locale/update/updates.properties");
			var downloadsBundle = _RestMan.strings.createBundle("chrome://mozapps/locale/downloads/downloads.properties");
			var questionMark = downloadsBundle.GetStringFromName("quitCancelDownloadsAlertTitle").match(/.$/); // ?
			_RestMan.askOnRestart = _RestMan.rmPrefBranch.getBoolPref("askOnRestart");
			_RestMan.restartMessage = updateBundle.formatStringFromName("restartButton", [_RestMan.appName], 1) + questionMark; // Restart brandShortName now
			_RestMan.neverAskLabel = askBundle.GetStringFromName("dontAskAgain");	// Don't ask me again
		};
	},

	winTabCounter: function RestMan_winTabCounter(aParam, aCaller) {
		// Functional module called when/where it is needed, to count windows & tabs
		// and check for private browsing mode, to avoid code repetitions.
		// It should be implemented in a separate jsm file for better code managing but
		// pm 1.5 & 2 don't support js modules..
		_RestMan.onLoadStrings();
		var callerName; // This function is also called by our component, which works in strict mode.
		// We get the caller of this function as follows, because the "function.caller" will not be
		// allowed in strict mode soon.
		if(aCaller == "rro") {
			callerName = "RestMan_RestartRequestObserver";
		} else if(aCaller == undefined) {
			callerName = RestMan_winTabCounter.caller.name;
		};
		if(_RestMan.codeMonitoring) {
			if(_RestMan.appVer >=20) {
				cu["import"]("resource://gre/modules/AddonManager.jsm", _RestMan); 	// We have to fool appVer 1.5, which can't
				// see in the future.. Otherwise it produces syntax error - cu.import("resource:...").
				_RestMan.AddonManager.getAddonByID("privateTab@infocatcher", function(pt) {
				if(pt && !pt.appDisabled && !pt.userDisabled && pt.isActive) { // "Private Tab" extension is installed & enabled.
					_RestMan.privateTabExt = true;
				} else {
					_RestMan.privateTabExt = false;
				};});
			};
			var parameter = aParam;
			if(aParam == undefined || aParam == "") {
				parameter = "none";
			};
			_RestMan.promptSrvc.alert(null, "winTabCounter caller:",
				callerName + "\n\n" + "Passed parameter: " + parameter);
		};
		// It is necessary to reset the values below on every call of this function,
		// in case this function has been called again previously in the same scope.
		_RestMan.allBrWinCount = 0;      // All browser windows.
		_RestMan.nonPrvWinCount = 0;     // All non-private browser windows.
		_RestMan.privWinCount = 0;       // All private browser windows.
		_RestMan.nPrWinVPrTbCnt = 0;     // Non-private windows with all their visible tabs private.
		_RestMan.nPrWnNprVTbCnt = 0;     // Non-private windows with at least 1 non-private visible tab.
		_RestMan.allTabCount = 0;        // All tabs.
		_RestMan.visTabCount = 0;        // All visible tabs.
		_RestMan.nonPrVisTabCnt = 0;     // All non-private visible tabs.
		_RestMan.prvVisTabCount = 0;     // All private visible tabs.
		_RestMan.allTabCountCW = 0;      // All tabs in current window.
		_RestMan.nPrVisTabCntCW = 0;     // Non-private visible tabs in current window.
		_RestMan.prvVisTabCntCW = 0;     // Private visible tabs in current window.
		// Note: Visible tabs = tabs in the current tab group.
		_RestMan.allWinPrivate = true;   //
		_RestMan.isWinPrivate = true;    // Current window.
		_RestMan.allTabsPrivate = false; //
		_RestMan.allTabsPrvCW = false;   // Current window.
		var allBrowserWinEnum = _RestMan.winMed.getEnumerator("navigator:browser");
		while(allBrowserWinEnum.hasMoreElements()) {
			_RestMan.allBrWinCount++;
			var eachBrowserWin = allBrowserWinEnum.getNext();
			if(_RestMan.appVer >=20) {
			// The js module "PrivateBrowsingUtils.jsm" is already imported by browser.js.
				if(eachBrowserWin.PrivateBrowsingUtils.isWindowPrivate(eachBrowserWin)) {
					_RestMan.privWinCount++;
				};
				if(!(eachBrowserWin.PrivateBrowsingUtils.isWindowPrivate(eachBrowserWin))) {
					_RestMan.allWinPrivate = false;
				};
			};
			var tabBrowser = eachBrowserWin.gBrowser; // window.gBrowser.tabs
			if(tabBrowser) {
				if(_RestMan.appVer >=20) {
					_RestMan.allTabCount += tabBrowser.browsers.length
										- tabBrowser._numPinnedTabs;
					_RestMan.visTabCount += tabBrowser.visibleTabs.length
										- tabBrowser._numPinnedTabs;
				} else if(_RestMan.appVer <20) {
					_RestMan.allTabCount += tabBrowser.browsers.length;
				};
			};
			if(_RestMan.appVer >=20) {
				Array.forEach(tabBrowser.visibleTabs, function(tabElement) {
					if(tabElement.hasAttribute("privateTab-isPrivate")) {
					// Compatibility with "Private Tab" extension.
						_RestMan.prvVisTabCount++;
					};
				});
				_RestMan.nonPrVisTabCnt = _RestMan.visTabCount - _RestMan.prvVisTabCount;
				if(!(eachBrowserWin.PrivateBrowsingUtils.isWindowPrivate(eachBrowserWin))) {
					// Compatibility with "Private Tab" extension. Detection of non-private
					// windows with all their tabs private.
					var prvVisTabCountInNonPrvWin = 0;
					var nonPrvVisTabCountInNonPrvWin = 0;
					Array.forEach(tabBrowser.visibleTabs, function(tabElement) {
						if(tabElement.hasAttribute("privateTab-isPrivate")) {
							prvVisTabCountInNonPrvWin++;
						};
						if(!(tabElement.hasAttribute("privateTab-isPrivate"))) {
							nonPrvVisTabCountInNonPrvWin++;
						};
					});
					if(nonPrvVisTabCountInNonPrvWin ==0 && prvVisTabCountInNonPrvWin >0) {
						_RestMan.nPrWinVPrTbCnt++; // Non-private windows with all their visible tabs private.
					};							   // Compatibility with "Private Tab" extension.
					if(nonPrvVisTabCountInNonPrvWin >=1) {
						_RestMan.nPrWnNprVTbCnt++; // Non-private windows with at least 1 non-private visible tab.
					};							   // Compatibility with "Private Tab" extension.
				};
			};
		};
		if(_RestMan.appVer >=20) {
			_RestMan.nonPrvWinCount = _RestMan.allBrWinCount - _RestMan.privWinCount;
			if(_RestMan.nonPrVisTabCnt ==0 && _RestMan.prvVisTabCount >0) {
				_RestMan.allTabsPrivate = true;
			};
		};
		var currentWinTabBrowser = window.gBrowser; // window.gBrowser.tabs
		if(currentWinTabBrowser) { // Count tabs / get privacy status of current window.
			if(_RestMan.appVer >=20) {
				if(!(window.PrivateBrowsingUtils.isWindowPrivate(window))) {
					_RestMan.isWinPrivate = false;
				};
				_RestMan.allTabCountCW += currentWinTabBrowser.browsers.length
									 - currentWinTabBrowser._numPinnedTabs;
				Array.forEach(currentWinTabBrowser.visibleTabs, function(tabElement) {
					if(!(tabElement.hasAttribute("privateTab-isPrivate"))) {
					// Compatibility with "Private Tab" extension.
						_RestMan.nPrVisTabCntCW++;
					};
				});
				Array.forEach(currentWinTabBrowser.visibleTabs, function(tabElement) {
					if(tabElement.hasAttribute("privateTab-isPrivate")) {
					// Compatibility with "Private Tab" extension.
						_RestMan.prvVisTabCntCW++;
					};
				});
				if(_RestMan.nPrVisTabCntCW ==0 && _RestMan.prvVisTabCntCW >0) {
					_RestMan.allTabsPrvCW = true;
				};
			} else if(_RestMan.appVer <20) {
				_RestMan.allTabCountCW += currentWinTabBrowser.browsers.length;
			};
		};
		if(_RestMan.codeMonitoring) {
			if(_RestMan.appVer >=20) {
				if(_RestMan.privateTabExt) { // "Private Tab" extension is installed & enabled.
					_RestMan.promptSrvc.alert(null, "winTabCounter:",
						"-- \"Private Tab\" extension is active --" +
						"\n--------------------------------------------" +
						"\nAll browser windows:------------- " + _RestMan.allBrWinCount +
						"\nNon-private windows:------------ " + _RestMan.nonPrvWinCount +
						"\nPrivate windows:------------------ " + _RestMan.privWinCount +
						"\nNonPrvWin/All VisTabs prv:------ " + _RestMan.nPrWinVPrTbCnt +
						"\nNon-prvWin/>=1NonPrv tabs: - " + _RestMan.nPrWnNprVTbCnt +
						"\nAll tabs:---------------------------- " + _RestMan.allTabCount +
						"\nVisible tabs:------------------------ " + _RestMan.visTabCount +
						"\nNon-private visible tabs:--------- " + _RestMan.nonPrVisTabCnt +
						"\nPrivate visible tabs:--------------- " + _RestMan.prvVisTabCount +
						"\nAll windows are private: --------- " + _RestMan.allWinPrivate +
						"\nAll tabs are private: -------------- " + _RestMan.allTabsPrivate +
						"\n|---------- Current window: ----------|" +
						"\nWindow is private:---------------- " + _RestMan.isWinPrivate +
						"\nAll tabs: ---------------------------- " + _RestMan.allTabCountCW +
						"\nNon-private visible tabs:--------- " + _RestMan.nPrVisTabCntCW +
						"\nPrivate visible tabs:--------------- " + _RestMan.prvVisTabCntCW +
						"\nAll tabs are private:--------------- " + _RestMan.allTabsPrvCW +
						"\n-----------------------------------------------------------" +
						"\nNote: Visible tabs = tabs in the current tab group."
					);
				} else {
					_RestMan.promptSrvc.alert(null, "winTabCounter:",
						"All browser windows:------ " + _RestMan.allBrWinCount +
						"\nNon-private windows:----- " + _RestMan.nonPrvWinCount +
						"\nPrivate windows:----------- " + _RestMan.privWinCount +
						"\nAll tabs:--------------------- " + _RestMan.allTabCount +
						"\nVisible tabs: ---------------- " + _RestMan.visTabCount +
						"\nAll windows are private: -- " + _RestMan.allWinPrivate +
						"\n|------- Current window: -------|" +
						"\nWindow is private: -------- " + _RestMan.isWinPrivate +
						"\nAll tabs:--------------------- " + _RestMan.allTabCountCW +
						"\nVisible tabs: ---------------- " + _RestMan.nPrVisTabCntCW +
						"\n-----------------------------------------------------------" +
						"\nNote: Visible tabs = tabs in the current tab group."
					);
				};
			} else if(_RestMan.appVer <20) {
				_RestMan.promptSrvc.alert(null, "winTabCounter:",
					"All browser windows:- " + _RestMan.allBrWinCount +
					"\nAll tabs: ---------------- " + _RestMan.allTabCount +
					"\n|--- Current window: ---|" +
					"\nAll tabs: ---------------- " + _RestMan.allTabCountCW
				);
			};
		};
		if(aParam == "tabCountCW") { // Current window, requester is "RestMan_onBrWinClose",
			// if the window is private or all of its tabs are private.
			if(_RestMan.detectOpenDlgs) {
				if(_RestMan.allBrWinCount == 1 && _RestMan.dialogDetector()) {
					_RestMan.openedDialog.focus();
					return;
				};
			};
			return _RestMan.allTabCountCW;
		};
		if(callerName == "RestMan_onBrWinClose" || // By pressing the close button.
			callerName == "oncommand") { // By Ctrl+Shift+w, "DOMWindowClose" event.
			if(_RestMan.detectOpenDlgs) {
				if(_RestMan.allBrWinCount == 1 && _RestMan.dialogDetector()) {
					_RestMan.openedDialog.focus();
					return;
				};
			};
			if(!_RestMan.isWinPrivate && !_RestMan.allTabsPrvCW && aParam == "checkPrivacy") {
				return true;
			};
			if(_RestMan.allBrWinCount == 1 && !_RestMan.isWinPrivate && !_RestMan.allTabsPrvCW && aParam == "brHandlesClosing") {
				_RestMan.onQuitRequest();
			} else if(_RestMan.allBrWinCount > 1 && (_RestMan.isWinPrivate || _RestMan.allTabsPrvCW) && aParam == "rmHandlesClosing") {
				_RestMan._RestManOnRestartRequest("close", "handleCloseWin");
			} else if(_RestMan.allBrWinCount == 1 && (_RestMan.isWinPrivate || _RestMan.allTabsPrvCW) && aParam == "rmHandlesClosing") {
				_RestMan._RestManOnRestartRequest("close", "handleCloseLastWin");
			};
		};
	},

	dialogDetector: function() {
		// We check if our restart or warning dialog or any other dialog of this
		// type or alert type or our options dialog or pm's prefs dialog or some
		// other dialog type is currently opened, to prevent showing multiple dialogs.
		// Also, some dialogs may need user's attention or intervention.
		// (Note: although most of these dialogs have a "modal" attribute, they can
		// be opened in different windows. So, multiple dialogs is possible.)
		// (Unfortunately, some dialogs, e.g. "commonDialog.xul" & "about.xul", don't have "windowtype" attribute..
		// Here we give them one temporarily, as long as they are loaded, if they are.
		// We do this also for some other child dialogs, opened over their parent dialog.)
		var allWinEnum = _RestMan.winMed.getEnumerator(null);
		while(allWinEnum.hasMoreElements()) {
			var eachWin = allWinEnum.getNext();
			var commonDlg = eachWin.document.getElementById("commonDialog");
			var genericAbout = eachWin.document.getElementById("genericAbout");
			var loaddevice = eachWin.document.getElementById("loaddevice");
			var setpass = eachWin.document.getElementById("set_password");
			var editCaCert = eachWin.document.getElementById("editCaCert");
			var delCert = eachWin.document.getElementById("deleteCertificate");
			var pwdedit = eachWin.document.getElementById("pwdedit"); // Password edit dialog of "Saved Password Editor" extension.
			var peOptions = eachWin.document.getElementById("PE_options"); // "Paste Email (original)"
			var pmcOptions = eachWin.document.getElementById("pmcommander-prefs"); // "Pale Moon Commander"
			var status4evarPrefs = eachWin.document.getElementById("status4evar-prefs");
			if(commonDlg) {
				commonDlg.setAttribute("windowtype", "dialog:commonDialog");
			};
			if(genericAbout) {
				genericAbout.setAttribute("windowtype", "dialog:genericAbout");
			};
			if(loaddevice) { // Child dialog.
				loaddevice.setAttribute("windowtype", "dialog:loaddevice");
			};
			if(setpass) { // Child dialog.
				setpass.setAttribute("windowtype", "dialog:setpass");
			};
			if(editCaCert) { // Child dialog.
				editCaCert.setAttribute("windowtype", "dialog:editCaCert");
			};
			if(delCert) { // Child dialog.
				delCert.setAttribute("windowtype", "dialog:delCert");
			};
			if(pwdedit) { // "Saved Password Editor" extension, edit dialog.
				pwdedit.setAttribute("windowtype", "spe:pwdedit");
			};
			if(peOptions) { // "Paste Email (original)" extension, options dialog.
				pwdedit.setAttribute("windowtype", "pasteemail:options");
			};
			if(pmcOptions) { // "Pale Moon Commander" extension, options dialog.
				pmcOptions.setAttribute("windowtype", "pmcommander:prefs");
			};
			if(status4evarPrefs) {
				status4evarPrefs.setAttribute("windowtype", "status4evar:prefs");
			};
		};
		if(_RestMan.detectOpenDlgs) {
			_RestMan.openedDialog = _RestMan.winMed.getMostRecentWindow("Browser:Preferences") || // Browser's pref window.
									_RestMan.winMed.getMostRecentWindow("dialog:commonDialog") ||
									_RestMan.winMed.getMostRecentWindow("dialog:genericAbout") ||
									_RestMan.winMed.getMostRecentWindow("alert:alert") || // ..?
									_RestMan.winMed.getMostRecentWindow("restman:options") || // Our options dialog.
									_RestMan.winMed.getMostRecentWindow("Shell:SetDesktopBackground") ||
									_RestMan.winMed.getMostRecentWindow("Browser:Permissions") ||
									_RestMan.winMed.getMostRecentWindow("dialog:editCaCert") ||
									_RestMan.winMed.getMostRecentWindow("dialog:delCert") ||
									_RestMan.winMed.getMostRecentWindow("mozilla:exceptiondialog") || // Child dialog.
									_RestMan.winMed.getMostRecentWindow("mozilla:certmanager") ||
									_RestMan.winMed.getMostRecentWindow("dialog:loaddevice") ||
									_RestMan.winMed.getMostRecentWindow("dialog:setpass") ||
									_RestMan.winMed.getMostRecentWindow("Toolkit:PasswordManagerExceptions") ||
									_RestMan.winMed.getMostRecentWindow("mozilla:devicemanager") ||
									_RestMan.winMed.getMostRecentWindow("spe:pwdedit") || // "Saved Password Editor" extension.
									_RestMan.winMed.getMostRecentWindow("Toolkit:PasswordManager") ||
									_RestMan.winMed.getMostRecentWindow("Addons:Compatibility") ||
									_RestMan.winMed.getMostRecentWindow("Addons:Install") ||
									_RestMan.winMed.getMostRecentWindow("pasteemail:options") || // "Paste Email (original)" extension.
									_RestMan.winMed.getMostRecentWindow("pasteemailplus:alert") || // "Paste Email Plus" extension.
									_RestMan.winMed.getMostRecentWindow("pasteemailplus:help") || //
									_RestMan.winMed.getMostRecentWindow("pasteemailplus:options") || //
									_RestMan.winMed.getMostRecentWindow("tiletabs-options") || // "Tile Tabs" extension.
									_RestMan.winMed.getMostRecentWindow("Browser:Confmania") || // "Configuration Mania" extension.
									_RestMan.winMed.getMostRecentWindow("pmcommander:prefs") || // "Pale Moon Commander" extension.
									_RestMan.winMed.getMostRecentWindow("status4evar:prefs");
		} else {
			_RestMan.openedDialog = _RestMan.winMed.getMostRecentWindow("restman:options"); // Always detection for presence of our options dialog.
		};
		if(_RestMan.openedDialog !== null) {
			return true;
		};
		return false;
	},

	sessionSaveHelper: function RestMan_sessionSaveHelper(aParam) {
	// In case of a mixed session (normal & private), if the last non-private window
	// is closed and thereafter or after a while pm is quitted or restarted, there
	// are some problems on next startup: some versions (e.g. pm 23) show the "about:sessionrestore"
	// page almost always, while some other (e.g. pm 27 & 28) cause the session of last non-private
	// window to be lost. Pm by design, saves the session of the last non-private window
	// and any other non-private data, only when the last private window is closed (not exit or restart pm).
	// We try here to simulate "closing of the last private window" on pm exit or restart, in order to
	// save the session of the last non-private window, on exit and restart too. It seems to work..  :-)
	// pm 3.5-19 manage very reliably the session, before and after the private browsing mode..
	// Note: This is questionable: "window not closed during a regular shut-down"
	// (SessionStore.jsm in pm 28, onClose(), line 899).
	// Why not to save any non-private data on shutdown (or even on restart) too? SessionStore.jsm filters and
	// excludes the private data, saving only the non-private data, so there are no any security or privacy concerns.
	// So, why not this procedure to not take place also on shutdown or restart? A good answer is wanted  :-)
		if(_RestMan.codeMonitoring) {
			var parameter = aParam;
			if(aParam == undefined) {
				parameter = "none";
			};
			_RestMan.promptSrvc.alert(null, "sessionSaveHelper caller:",
				RestMan_sessionSaveHelper.caller.name // arguments.callee.caller
				+ "\n\n" + "Passed parameter: " + parameter);
		};
		var sshCaller;
		if(RestMan_sessionSaveHelper.caller.name == "doRestartNoRestoreCmd") {
			sshCaller = "rnrCmd"; // The caller will not be available later on, when we'll need to know who is.
		};
		var currentBrWindow = _RestMan.winMed.getMostRecentWindow("navigator:browser");
		currentBrWindow.addEventListener("unload", function ssHelperOnUnload() { // This is the last window that will be closed.
			this.removeEventListener("unload", ssHelperOnUnload, false);
			if(_RestMan.codeMonitoring &&
				!_RestMan.restmanSrvc.getPrvBrowsingAutostart()
				&& _RestMan.allWinPrivate) {
				_RestMan.promptSrvc.alert(null, "sessionSaveHelper:",
										  "The last private browser window is unloaded." +
										  "\nThe session of the last non-private window has been saved.");
			};
			if(sshCaller == "rnrCmd") {
				_RestMan._RestManOnRestartRequest("restart", "rnrCmd");
			};
		}, false);
		if(aParam == "handleExit" || aParam == "restart") {
			var allWinEnum = _RestMan.winMed.getEnumerator(null);
			while(allWinEnum.hasMoreElements()) {
				var winToBeClosed = allWinEnum.getNext();
				if (winToBeClosed != currentBrWindow) {
					winToBeClosed.close(); // Close all windows except the current.
				};
			};
		};
		// currentBrWindow.close();
		// We prefer to let browser manage the last window closing, to do first its default
		// checks, instead of a "cooldown" window.close(). 
		// First, we close all tabs but one, to prevent pm of showing its warning dialog,
		// if applicable and if tabs >1, on our restart or exit process. We show our own warning dialog
		// (optional) specifically in private browsing mode, because pm has only one option to
		// show/hide its warning dialog on closing multiple tabs, which applies in both private
		// and non-private browsing mode.
		// In pm >=26, there is a warning dialog on exiting or closing the last window in private
		// browsing mode, but as above, its option to hide it is applied also on the quit dialog
		// in non-private browsing mode.
		var tabBrowser = currentBrWindow.gBrowser;
		var tabs = tabBrowser.visibleTabs;
		var sTab = tabBrowser.selectedTab;
		for(var i = tabs.length - 1; i >= 0; --i) { // Close all tabs except the current.
			if (tabs[i] != sTab) {
				tabBrowser.removeTab(tabs[i], {animate: false});
			};
		};
		currentBrWindow.setTimeout(function() { // Ffx 34 nightly, on exit & on close last win, shows its warning dialog.
		// It's not logical.. This small delay fixes this issue. We leave it as is for pm too, since it's not noticeable and doesn't harm. Just in case..
			currentBrWindow.BrowserTryToCloseWindow(); // browser.js
		}, 100);
	},

	_RestManOnRestartRequest: function RestMan_RestManOnRestartRequest(aCallReason, aParam, aCaller) {
/* Our restart logic according to user actions, preferences and pm versions:
	case1(default):							If visible tabs >1, will be shown the restart dialog,
											else no dialog, tab will be restored after restart.
	case2(!askOnRestart && sessionRestore): No dialog, the session will be restored after restart.
	case3(!askOnRestart && !sessionRestore && warnOnCloseMT): If visible tabs >1, will be shown
											our "Warn me when closing multiple tabs" dialog, else
											no dialog, tab will not be restored after restart.
	case4(!askOnRestart && !sessionRestore && !warnOnCloseMT): No dialogs, the session
											will not be restored after restart.
	case5: In private browsing mode, if visible tabs >1, will be shown only our "Warn me..."
			dialog, on app restart, exit or window closing, if this has been selected in options. */

		var callerName; // This function is called also by our component, which works in strict mode.
		// We get the caller of this function as follows, because the "function.caller" will not be
		// allowed in strict mode soon.
		if(aCaller == "rro") {
			callerName = "RestMan_RestartRequestObserver";
		} else if(aCaller == undefined) {
			if(RestMan_RestManOnRestartRequest.caller !== null) {
				callerName = RestMan_RestManOnRestartRequest.caller.name;
			} else {
				callerName = "top level!";
			};
		};
		var codMonPref = _RestMan.rmPrefBranch.getBoolPref("alertsEnabled");
		if(codMonPref) { // "_RestMan.codeMonitoring" is not available here in some specific case,
		// where we have closed the window and we can't call "_RestMan.onLoadStrings()".
		// If we do it, the code will be broken at this point.
			var parameter2 = aParam;
			if(aParam == undefined || aParam == "") {
				parameter2 = "none";
			};
			_RestMan.promptSrvc.alert(null, "_RestManOnRestartRequest caller:",
				callerName + "\n\nPassed parameters:" + "\nCall reason: " +
				aCallReason + "\nParameter: " + parameter2);
		};
		if(_RestMan.detectOpenDlgs) {
			if(_RestMan.dialogDetector() && aCallReason == "restart") {
				_RestMan.openedDialog.focus();
				return;
			};
		};
		var cancelQuit = cc["@mozilla.org/supports-PRBool;1"].createInstance(ci.nsISupportsPRBool);
		if(aParam == "rnrCmd") {
			doRestartNoRestoreCmd();
			return;
		} else if(aParam == "doRestart") {
			doRestartCmd();
			return;
		} else if(aParam == "doRestartNoRestore") {
			doRestartNoRestore(aParam);
			return;
		};
		_RestMan.sharedStrings();
		var privateBrModeFlag = false; // Used by warning dialog, to manage the correct preference.
		var warningDlgFlag = false; // Used to determine if the warning dialog will be shown.
		// No restart dialog if there are no at least 2 non-private tabs in non-private window(s)
		// (excluding pinned tabs).
		// In private & non-private or mixed browsing mode, we show also the warning dialog,
		// if this has been selected in options ("Warn me when closing multiple tabs").
		// The "privateBrModeFlag" is used to determine the browsing mode, to change the
		// corresponding preference if the check box of the warning dialog is unchecked.
		// It is used also as an indicator, to perform our special actions on pm restarting/exiting
		// from within private browsing mode.
		_RestMan.winTabCounter();
		if(_RestMan.appVer >=20) {
			function codeMon() {
				var parameter = aParam;
				if(aParam == undefined) {
					parameter = "none";
				};
				if(_RestMan.codeMonitoring) {
					_RestMan.promptSrvc.alert(null, "Private browsing mode flags:",
						"privateBrModeFlag: " + privateBrModeFlag + 
						"\nwarningDlgFlag: " + warningDlgFlag +
						"\n(Call reason: " + aCallReason + ")" +
						"\n(Parameter: " + parameter + ")");
				};
			};
			if((_RestMan.allWinPrivate || _RestMan.allTabsPrivate) && aCallReason !== "close") {
					privateBrModeFlag = true;
				if(!_RestMan.warnOnCloseMPT) { // multiple private tabs.
					if(aCallReason == "restart") { // Passed by our ui restart items or our observer.
						codeMon();
						doRestartNoRestoreCmd("ssHelper");
						return;
					} else if(aCallReason == "exit") { // On "File --> Exit", private browsing mode.
						codeMon();					   // Passed by our observer.
						_RestMan.onQuitRequest("handleExit");
						return;
					};
				} else if(_RestMan.warnOnCloseMPT) { // multiple private tabs.
					warningDlgFlag = true;
					if(aCallReason == "restart") { // Passed by our ui restart items or our observer.
						codeMon();
						doRestartNoRestore(aCallReason);
						return;
					} else if(aCallReason == "exit") { // On "File --> Exit", private browsing mode.
						codeMon();					   // Passed by our observer.
						doRestartNoRestore("handleExit");
						return;
					};
				};
			};
			if(callerName == "RestMan_winTabCounter" && aCallReason == "close") {
				// We are here if the window to be closed is private or all of its tabs are private
				// or it's the last browser window which is private or all of its tabs are private.
				privateBrModeFlag = true;
				if(_RestMan.warnOnCloseMPT) { // multiple private tabs.
					warningDlgFlag = true;
				};
				codeMon();
				doRestartNoRestore(aParam); // aParam will be "handleCloseWin" or "handleCloseLastWin".
				return;
			};
		};
		if(aCallReason == "restart") { // Passed by our ui restart items or our observer.
		// Non-private browsing mode, all supported app versions.
			if(_RestMan.askOnRestart) {
				showRestartDlg();
			} else if(_RestMan.appVer >=20) {
				if(!_RestMan.askOnRestart && _RestMan.sessionRestore) {
					doRestartCmd();
				} else if(!_RestMan.askOnRestart && !_RestMan.sessionRestore
					&& _RestMan.warnOnCloseMT) { // multiple tabs.
					warningDlgFlag = true;
					if(_RestMan.codeMonitoring) {
						_RestMan.promptSrvc.alert(null, "Non-private browsing mode flag:",
							"warningDlgFlag: " + warningDlgFlag);
					};
					doRestartNoRestore("doRestartNoRestore");
				} else if(!_RestMan.askOnRestart && !_RestMan.sessionRestore
					&& !_RestMan.warnOnCloseMT) { // multiple tabs.
					doRestartNoRestore("doRestartNoRestore");
				};
			} else if(_RestMan.appVer <20) {
				doRestartCmd();
			};
		};
		function doRestartCmd() {
			// Session will be restored after restart (except the part of the session which is private).
			// (Pm 1.5 can't restore the session by any means..)
			if(_RestMan.codeMonitoring) {
				_RestMan.promptSrvc.alert(null, "doRestartCmd caller:",
					doRestartCmd.caller.name); // arguments.callee.caller
			};
			if(_RestMan.appVer >=3) {
				_RestMan.restmanSrvc.observerUnregister();
			};
			_RestMan.obsSrvc.notifyObservers(cancelQuit, "quit-application-requested", "restart");
			if(cancelQuit.data) { // The quit request has been canceled.
				if(_RestMan.appVer >=3) {
					_RestMan.restmanSrvc.observerRegister();
				};
				return false;
			};
			if(_RestMan.appVer >=20 && _RestMan.restoreUserSet) {
				_RestMan.onQuitRequest("restoreUserSetCmd");
			};
			if(_RestMan.appVer >=2) {
				_RestMan.brPrefBranch.setBoolPref("sessionstore.resume_session_once", true);
			};
			if(_RestMan.invalCaches) {
				if(_RestMan.appVer >=20) {
					_RestMan.appInfo.QueryInterface(ci.nsIXULRuntime).invalidateCachesOnRestart();
				} else if(_RestMan.appVer <3) { // For pm 3-3.6, this is performed in "restartDlg3to19".
					var cacheFiles = ["compreg.dat", "xpti.dat", "extensions.cache", "extensions.ini"]; // "extensions.rdf", "pluginreg.dat"
					cacheFiles.forEach(function(name) {
						var profDirFiles = _RestMan.dirSrvc.get("ProfD", ci.nsIFile);
						if(profDirFiles) {
							profDirFiles.append(name);
						};
						if(profDirFiles.exists()) {
							profDirFiles.remove(false);
						};
					});
				};
			};
			_RestMan.appStartup.quit(_RestMan.appStartup.eAttemptQuit | _RestMan.appStartup.eRestart);
		};
		function doRestartNoRestoreCmd(aParam) {
			// Non-private session will be retained for possible desirable "Restore Previous Session" action,
			// but will not be restored after restart.
			if(_RestMan.codeMonitoring) {
				var parameter = aParam;
				if(aParam == undefined) {
					parameter = "none";
				};
				_RestMan.promptSrvc.alert(null, "doRestartNoRestoreCmd caller:",
					doRestartNoRestoreCmd.caller.name // arguments.callee.caller
					+ "\n\n" + "Passed parameter: " + parameter);
			};
			if(aParam == "ssHelper") {
				_RestMan.sessionSaveHelper("restart");
				return;
			};
			_RestMan.restmanSrvc.observerUnregister();
			_RestMan.obsSrvc.notifyObservers(cancelQuit, "quit-application-requested", "restart");
			if(cancelQuit.data) { // The quit request has been canceled.
				_RestMan.restmanSrvc.observerRegister();
				return false;
			};
			// If "browser.startup.page" = 3, the session will be restored, even if
			// "sessionstore.resume_session_once = false". This does not serve us here..
			// We check this value, change it if necessary and restore it on "restart & restore"
			// or on quit. Also, maybe it is needed to set "warnOnQuit = false".
			// All the details for our settings restoration process, can be found below, in "onQuitRequest" function.
			if(aParam == "checkStartupType") { // This has a meaning only in non-private sessions.
			// In case of a mixed session, the private part will be lost as usual.
				var value = _RestMan.brPrefBranch.getIntPref("startup.page");
				if(value == 3) {
					_RestMan.brPrefBranch.setIntPref("startup.page", 1); // pm's default.
					_RestMan.rmPrefBranch.setBoolPref("restore_UserSettings", true);
					if(_RestMan.brPrefBranch.getBoolPref("warnOnQuit")) { // pm's default.
						_RestMan.brPrefBranch.setBoolPref("warnOnQuit", false);
						_RestMan.rmPrefBranch.setBoolPref("restore_warnOnQuitPref", true);
					};
				};
			};
			if(_RestMan.invalCaches) {
				_RestMan.appInfo.QueryInterface(ci.nsIXULRuntime).invalidateCachesOnRestart();
			};
			_RestMan.appStartup.quit(_RestMan.appStartup.eAttemptQuit | _RestMan.appStartup.eRestart);
		};
		function doRestartNoRestore(aParam) {
			// Session will not be restored after restart, but will be saved, in case the 
			// user wants to "Restore Previous Session" (excluding a private session or 
			// the part of the session which is private).
			// Session's state will be saved just before restart, for the case of a change
			// and thereafter restart, before the expiration of current session save interval
			//(default interval = 15secs).
			var parameter = aParam;
			if(aParam == undefined) {
				parameter = "none";
			};
			if(_RestMan.codeMonitoring) {
				_RestMan.promptSrvc.alert(null, "doRestartNoRestore caller:",
					doRestartNoRestore.caller.name // arguments.callee.caller
					+ "\n\n" + "Passed parameter: " + parameter);
			};
			var doRestartNoRestorePrefObserver = {
			// We consider the use of a pref observer as more reliable and app friendly,
			// instead of locking the pref of interest.
				pref2Observe: "sessionstore.resume_session_once",
				observe: function(aSubject, aTopic, aData) {
					if(_RestMan.codeMonitoring) {
						_RestMan.promptSrvc.alert(null, "doRestartNoRestore observe:",
							"Subject: " + aSubject + "\nTopic: " +	aTopic + "\nData: " +aData);
					};
					if(aTopic == "nsPref:changed" && aData == this.pref2Observe) {
						if(_RestMan.brPrefBranch.getBoolPref(this.pref2Observe)) {
							_RestMan.brPrefBranch.setBoolPref(this.pref2Observe, false);
						};
					};
				},
				observerRegister: function() {
					if(_RestMan.brPrefBranch.getBoolPref(this.pref2Observe)) {
						_RestMan.brPrefBranch.setBoolPref(this.pref2Observe, false);
					};
					if(!_RestMan.dnsPrefObserver) {
						_RestMan.brPrefBranch.addObserver(this.pref2Observe, this, false);
						_RestMan.dnsPrefObserver = true;
						if(_RestMan.codeMonitoring) {
							_RestMan.promptSrvc.alert(null, "doRestartNoRestore pref observer:", "Registered");
						};
					};
				},
				observerUnregister: function() {
					if(_RestMan.dnsPrefObserver) {
						_RestMan.brPrefBranch.removeObserver(this.pref2Observe, this);
						_RestMan.dnsPrefObserver = false;
						if(_RestMan.codeMonitoring) {
							_RestMan.promptSrvc.alert(null, "doRestartNoRestore pref observer:", "Unregistered");
						};
					};
				}
			};
			if(!_RestMan.allWinPrivate && !_RestMan.allTabsPrivate && aParam == "doRestartNoRestore") { // Compatibility
			// with "Private Tab" extension. We may have non-private windows but all of their tabs may be private.
				doRestartNoRestorePrefObserver.observerRegister();
			};
			if(privateBrModeFlag) {
				if(_RestMan.allTabCount<2 || !warningDlgFlag) {
					if(aParam == "handleCloseWin") {
						this.close();
						return;
					} else if(aParam == "handleCloseLastWin" || aParam == "handleExit") {
						_RestMan.onQuitRequest(aParam);
						return;
					} else if(aParam == "restart") {
						doRestartNoRestoreCmd("ssHelper");
						return;
					};
				} else if(warningDlgFlag && 
					(aParam == "restart" || aParam == "handleCloseWin" ||
					 aParam == "handleCloseLastWin" || aParam == "handleExit")) {
						warningDialog(aParam);
						return;
				};
			} else {
				if(_RestMan.visTabCount<2 || !warningDlgFlag) {
					doRestartNoRestoreCmd("checkStartupType");
					return;
				} else {
					warningDialog(aParam)
				};
			};
			function warningDialog(aParam) { // This is our warning dialog, used in four different cases,
			// with the corresponding title and message: 1 - On "restart no restore", 2 - on closing a private window
			// with >1 tabs, 3 - on closing the last browser window if it is private with >1 tabs and 4 - on exiting
			// pm from within pivate browsing mode, if there are more than one tabs.
				if(_RestMan.codeMonitoring) {
					_RestMan.promptSrvc.alert(null, "warningDialog caller:",
						warningDialog.caller.name // arguments.callee.caller
						+ "\n\n" + "Passed parameter: " + parameter);
				};
				/*<Localized strings, used in warning dialog.>*********/
				var tabbrBundle = _RestMan.strings.createBundle("chrome://browser/locale/tabbrowser.properties");
				var quitBundle = _RestMan.strings.createBundle("chrome://browser/locale/quitDialog.properties");
				var warnDlgTitle;
				if(aParam == "restart" || aParam == "doRestartNoRestore") {
					warnDlgTitle = _RestMan.restartTitle + " - " + tabbrBundle.GetStringFromName("tabs.closeWarningTitle"); // Restart brandShortName - Confirm close
				} else if(aParam == "handleExit" // Passed by our js component on "File --> Exit".
					 || aParam == "handleCloseLastWin") {
					warnDlgTitle = quitBundle.formatStringFromName("quitDialogTitle", [_RestMan.appName], 1) + " - " +
									_RestMan.confirmLabel; // Quit brandShortName - Confirm
				} else if(aParam == "handleCloseWin") {
					warnDlgTitle = _RestMan.winMed.getMostRecentWindow("navigator:browser").document.getElementById("closeWinLabel").value // Close Window
									+ " - " + _RestMan.confirmLabel; // Confirm (--> Close Window - Confirm)
				};
				var text1 = _RestMan.winMed.getMostRecentWindow("navigator:browser").document.getElementById("privateBrowsingTitle").value.replace(/[()]/g,""); // Private Browsing
				var tabsCount = 0;
				if(aParam == "restart" || aParam == "doRestartNoRestore" || aParam == "handleExit") {
					tabsCount = _RestMan.visTabCount;
				} else if(aParam == "handleCloseWin" || aParam == "handleCloseLastWin") {
					tabsCount = _RestMan.allTabCountCW; // All tabs in current window.
				};
				var text2 = tabbrBundle.formatStringFromName("tabs.closeWarningMultipleTabs", [tabsCount], 1); // "You are about to close xx tabs"
				if(_RestMan.appVer >=26) { // We welcome and use the new string for the message in quit pm or
				// close last window dialog, in private browsing mode, in pm >=26, inherited from ffx 34 ("browser.showQuitWarning" = true).
				// However, moz-devs made for second time the same mistake, as in the restart dialog in ffx 3-19:
				// This message is a statement, not a question, while the label in the checkbox to hide the dialog says "Do not ask next time"!
				// When we use this message in our warning dialog, we combine it with the "Warn me..." label in the checkbox..
					var text3 = quitBundle.formatStringFromName("messagePrivate", [_RestMan.appName], 1); // "You're in private browsing mode. Quitting..."
					var openedTabsLabel = _RestMan.strings.createBundle("chrome://browser/locale/syncQuota.properties")
						.GetStringFromName("collection.tabs.label"); // Tabs
				};
				var warnMe = "&"+tabbrBundle.GetStringFromName("tabs.closeWarningPromptMe"); // "Warn me when..."
				var button0Title = tabbrBundle.GetStringFromName("tabs.closeButtonMultiple"); // Close tabs
				var _dlgMessage;
				var _defaultButton;
					if(privateBrModeFlag) {
						_defaultButton = _RestMan.promptSrvc.BUTTON_POS_1_DEFAULT;
						if(_RestMan.appVer >=26) {
							if(aParam == "restart" || aParam == "handleExit" || aParam == "handleCloseLastWin") {
								_dlgMessage = text1 + "\n" + "(" + openedTabsLabel + ": " + tabsCount + ")" + "\n" + text3;
							} else if(aParam == "handleCloseWin") {
								_dlgMessage = text1 + "\n\n" + text2;
							};
						} else if(_RestMan.appVer <26) {
							_dlgMessage = text1 + "\n\n" + text2;
						};
					} else {
						_dlgMessage = text2;
						_defaultButton = _RestMan.promptSrvc.BUTTON_POS_0_DEFAULT;
					};
				/*********************************</Localized strings>*/
				var Chkbx = {value: true};
				var flags = _RestMan.promptSrvc.BUTTON_TITLE_IS_STRING * _RestMan.promptSrvc.BUTTON_POS_0 +
							_RestMan.promptSrvc.BUTTON_TITLE_CANCEL    * _RestMan.promptSrvc.BUTTON_POS_1 +
							_defaultButton;
				var mostRecentBrowserWindow = _RestMan.winMed.getMostRecentWindow("navigator:browser");
				var userButtonChoice = _RestMan.promptSrvc.confirmEx(
						mostRecentBrowserWindow, warnDlgTitle, _dlgMessage,
						flags, button0Title, null, null, warnMe, Chkbx
						);
				switch (userButtonChoice) {
					case 1: // Cancel close or restart, close dialog.
						if(_RestMan.dnsPrefObserver) {
							doRestartNoRestorePrefObserver.observerUnregister();
						};	
						if(privateBrModeFlag) {
							privateBrModeFlag = false;
						};
						warningDlgFlag = false;
						break;
					case 0: // Close tabs/windows & restart or exit.
						if(!Chkbx.value) {
							if(!privateBrModeFlag) {
								_RestMan.rmPrefBranch.setBoolPref("warnOnClose_multipleTabs", false);
							} else {
								_RestMan.rmPrefBranch.setBoolPref("warnOnClose_multiplePrvTabs", false);
							};
						};
						if(privateBrModeFlag && aParam == "restart") {
							doRestartNoRestoreCmd("ssHelper");
						} else if(privateBrModeFlag && (aParam == "handleExit" || aParam == "handleCloseLastWin")) {
							_RestMan.onQuitRequest(aParam);
						} else if(privateBrModeFlag && aParam == "handleCloseWin") {
							this.close();
						} else if(!privateBrModeFlag && aParam == "doRestartNoRestore") {
							doRestartNoRestoreCmd("checkStartupType");
						};
						break;
				};
			};
		};
		function showRestartDlg() {
			if(_RestMan.appVer >=20) {
				restartDlg20plus();
			} else if(_RestMan.appVer <20 && _RestMan.appVer >=3) {
				restartDlg3to19();
			} else if(_RestMan.appVer <3) {
				restartDlg1v5to2();
			};
		};
		function restartDlg20plus() {
			/*  Pm >=20 don't have a restart dialog anymore..
				Our restart dialog here is based on the quit dialog.
			********************************************************/
			if(_RestMan.nonPrVisTabCnt<2) {
				RestMan_RestManOnRestartRequest("restart", "doRestart");
				return;
			};
			/*<Localized strings, used in restart dialog.>*********/
			var cdBundle = _RestMan.strings.createBundle("chrome://global/locale/commonDialogs.properties");
			var prefsBundle = _RestMan.strings.createBundle("chrome://browser/locale/preferences/preferences.properties");
			var quitBundle = _RestMan.strings.createBundle("chrome://browser/locale/quitDialog.properties");
			var restartDlgTitle = _RestMan.restartTitle + (" - ") + _RestMan.confirmLabel; // Restart brandShortName - Confirm
			var browserBundle = _RestMan.strings.createBundle("chrome://browser/locale/browser.properties");
			var dlgMessage;
			// Pm shows the "message" message even when, while there are at least 2 windows,
			// only 1 of them is non-private. However, private windows can not survive after quiting
			// or restarting..
			// We show the "message" message only when there are at least 2 non-private windows, with
			// at least 1 non-private tab in each one (Compatibility with "Private Tab" extension).
			if(_RestMan.nPrWnNprVTbCnt == 1) { // non-private window with non-private visible tabs.
				dlgMessage = quitBundle.formatStringFromName("messageNoWindows", [_RestMan.appName], 1);
			} else {
				dlgMessage = quitBundle.formatStringFromName("message", [_RestMan.appName], 1);
			};
			var saveAndquit = quitBundle.GetStringFromName("saveTitle"); // &Save and Quit
			var rep = /(\S+)(\s)(\S+)(\s)(\S+)/;
			var saveAnd = saveAndquit.replace(rep, "$1$2$3"); // &Save and Quit --> &Save and
			var restartBtn = browserBundle.GetStringFromName("safeModeRestartButton");
			/*********************************</Localized strings>*/

			var button0Title = saveAnd + (" ") + restartBtn.toLocaleLowerCase(); // &Save and restart
			var button1Title = quitBundle.GetStringFromName("cancelTitle"); // &Cancel
			var button2Title = "&Restart";
			if(!(/\(&.\)/).test(saveAndquit)) { // Probably it's not an eastern locale.
				button2Title = "&" + restartBtn; // .replace(restartAkey, "&"+restartAkey); // Restart --> &Restart
			} else {
				button2Title = restartBtn + "(&R)";
			};
			var neverAskChkbx = {value:false};
			var flags = _RestMan.promptSrvc.BUTTON_TITLE_IS_STRING * _RestMan.promptSrvc.BUTTON_POS_0 +
						_RestMan.promptSrvc.BUTTON_TITLE_IS_STRING * _RestMan.promptSrvc.BUTTON_POS_1 +
						_RestMan.promptSrvc.BUTTON_TITLE_IS_STRING * _RestMan.promptSrvc.BUTTON_POS_2 +
						_RestMan.promptSrvc.BUTTON_POS_0_DEFAULT;

			// We show the restart dialog always in non-private window, to avoid any possible
			// confusions, making the user think that private windows/tabs can be restored after restart.
			// Don't like that much the method below, but anyway it works..
			// A better way to get the most recent non-private browser window, is wanted :-)
			var mostRecentNonPrvBrowserWindow = {
				checkWinPrivacy: function() {
					var allBrowserWinEnum = _RestMan.winMed.getZOrderDOMWindowEnumerator("navigator:browser", true);
					while(allBrowserWinEnum.hasMoreElements()) {
						var browserWin = allBrowserWinEnum.getNext();
						var nonPrvVisTabCount = 0;
						Array.forEach(browserWin.gBrowser.visibleTabs, function(tabElement) {
							if(!(tabElement.hasAttribute("privateTab-isPrivate"))) {
							// Compatibility with "Private Tab" extension.
								nonPrvVisTabCount++;
							};
						});
						if(!(browserWin.PrivateBrowsingUtils.isWindowPrivate(browserWin)) && nonPrvVisTabCount >= 1) {
							return browserWin.focus();
						};
					};
				}
			};
			var userButtonChoice =
			_RestMan.promptSrvc.confirmEx(mostRecentNonPrvBrowserWindow.checkWinPrivacy(), restartDlgTitle,
										  dlgMessage, flags, button0Title, button1Title, button2Title,
										  _RestMan.neverAskLabel, neverAskChkbx
										  );
			switch (userButtonChoice) {
				case 2: // Save session & restart, no session restore.
					if(neverAskChkbx.value) {
						_RestMan.rmPrefBranch.setBoolPref("askOnRestart", false);
						_RestMan.rmPrefBranch.setBoolPref("restoreSession_onRestart", false);
						_RestMan.rmPrefBranch.setBoolPref("warnOnClose_multipleTabs", true);
					};
					RestMan_RestManOnRestartRequest("restart", "doRestartNoRestore");
					break;
				case 1: // Do nothing, close dialog.
					break;
				case 0: // Save session, restart & restore session.
					if(neverAskChkbx.value) {
						_RestMan.rmPrefBranch.setBoolPref("askOnRestart", false);
						_RestMan.rmPrefBranch.setBoolPref("restoreSession_onRestart", true);
					};
					RestMan_RestManOnRestartRequest("restart", "doRestart");
					break;
			};
		};
		function restartDlg3to19() {
			// In  pm 3-19, we call the default restart dialog of these versions.
			// Nothing special here, other than that we patch the message of the dialog,
			// adding a string in the end, to make it be a question to match with
			// "Do not ask next time" label in its check box (_RestManOnLoad function).
			// Moz-devs should do something like this, instead of removing this dialog..
			_RestMan.obsSrvc.notifyObservers(cancelQuit, "quit-application-requested", "restart");
			if(cancelQuit.data) {
				return false; // The quit request has been canceled.
			};
			_RestMan.restmanSrvc.observerUnregister();
			_RestMan.brPrefBranch.setBoolPref("sessionstore.resume_session_once", true);
			if(_RestMan.invalCaches && _RestMan.appVer <4) {
				var cacheFiles = ["compreg.dat", "xpti.dat", "extensions.cache", "extensions.ini"]; // "extensions.rdf", "pluginreg.dat"
				cacheFiles.forEach(function(name) {
					var profDirFiles = _RestMan.dirSrvc.get("ProfD", ci.nsIFile);
					if(profDirFiles) {
						profDirFiles.append(name);
					};
					if(profDirFiles.exists()) {
						profDirFiles.remove(false);
					};
				});
			};
			_RestMan.appStartup.quit(_RestMan.appStartup.eAttemptQuit | _RestMan.appStartup.eRestart);
		};
		function restartDlg1v5to2() {
			// A simple restart dialog for pm 1.5 & 2, similar to this one in pm 3-19.
			if(_RestMan.allTabCount<2) {
				RestMan_RestManOnRestartRequest("restart", "doRestart");
			} else {
				var dlgMessage = _RestMan.restartMessage;
				if(_RestMan.appVer <2) {
				// Pm 1.5 can't restore the session after restart. For this, we show also the message
				// "You are about to close xx open tabs."
					var tabsBundle = _RestMan.strings.createBundle("chrome://global/locale/tabbrowser.properties");
					var closeTabsMsg =  tabsBundle.formatStringFromName("tabs.closeWarningMultiple",
						[_RestMan.allTabCount], 1).match(/^[^\.\,\?\!]+[\.\,\?\!]/); // Does not work in some eastern languages..
					if(closeTabsMsg != null) {
						dlgMessage = closeTabsMsg + "\n\n" + _RestMan.restartMessage + "\n";
					};
				};
				var neverAskChkbx = {value: false};
				var flags = _RestMan.promptSrvc.BUTTON_TITLE_YES    * _RestMan.promptSrvc.BUTTON_POS_0 +
							_RestMan.promptSrvc.BUTTON_TITLE_CANCEL * _RestMan.promptSrvc.BUTTON_POS_1 +
							_RestMan.promptSrvc.BUTTON_POS_0_DEFAULT;
				var mostRecentBrowserWindow = _RestMan.winMed.getMostRecentWindow("navigator:browser");
				var userButtonChoice = _RestMan.promptSrvc.confirmEx(
					mostRecentBrowserWindow, null, dlgMessage, flags,
					null, null, null, _RestMan.neverAskLabel, neverAskChkbx
					);
				switch (userButtonChoice) {
					case 1: // Do nothing, close dialog.
						break;
					case 0: // Restart.
						if(neverAskChkbx.value) {
							_RestMan.rmPrefBranch.setBoolPref("askOnRestart", false);
						};
						RestMan_RestManOnRestartRequest("restart", "doRestart");
						break;
				};
			};
		};
	},

	// Prompt user to restart the browser in safe mode (Modified mozilla code from browser.js,
	// to prevent us to be engaged in the process of restart request/command for safe mode.).
	safeModeRestart: function() {
		// prompt the user to confirm
		var promptTitle = gNavigatorBundle.getString("safeModeRestartPromptTitle");
		var promptMessage = gNavigatorBundle.getString("safeModeRestartPromptMessage");
		var restartText = gNavigatorBundle.getString("safeModeRestartButton");
		var buttonFlags = _RestMan.promptSrvc.BUTTON_TITLE_IS_STRING * _RestMan.promptSrvc.BUTTON_POS_0 +
						  _RestMan.promptSrvc.BUTTON_TITLE_CANCEL    * _RestMan.promptSrvc.BUTTON_POS_1 +
						  _RestMan.promptSrvc.BUTTON_POS_0_DEFAULT;
		var rv = _RestMan.promptSrvc.confirmEx(window, promptTitle, promptMessage,
												buttonFlags, restartText, null, null,
												null, {});
		if(rv != 0) {
			return;
		} else {
			var cancelQuit = cc["@mozilla.org/supports-PRBool;1"].createInstance(ci.nsISupportsPRBool);
			_RestMan.restmanSrvc.observerUnregister(); // The actual modification in original mozilla code.
			_RestMan.obsSrvc.notifyObservers(cancelQuit, "quit-application-requested", "restart");
			if(!cancelQuit.data) {
				_RestMan.appStartup.restartInSafeMode(_RestMan.appStartup.eAttemptQuit);
			};
		};
	},

	onBrWinClose: function RestMan_onBrWinClose(event) {
		if(_RestMan.winTabCounter("checkPrivacy")) {
			// (In case of warning dialog, "WindowIsClosing()" (browser.js) function returns true only if
			// the dialog is accepted)
			if(!WindowIsClosing()) {
				return false;
			} else {
				_RestMan.winTabCounter("brHandlesClosing");
				return true;
			};
		} else {
			var tabCountCW = _RestMan.winTabCounter("tabCountCW"); // Tabs of current private window.
			if(tabCountCW == 1) {
				return true;
			} else if(tabCountCW > 1) {
				event.preventDefault();
				window.focus();
				_RestMan.winTabCounter("rmHandlesClosing");
				return;
			};
		};
		return false;
	},	

	onQuitRequest: function RestMan_onQuitRequest(aParam, aCaller) {
		_RestMan.sharedStrings();
		var parameter = aParam;
		if(aParam == undefined || aParam == "") {
			parameter = "none";
		};
		var callerName; // This function is called also by our component, which works in strict mode.
		// We get the caller of this function as follows, because the "function.caller" will not be
		// allowed in strict mode soon.
		if(aCaller == "rro") {
			callerName = "RestMan_RestartRequestObserver";
		} else if(aCaller == undefined) {
			callerName = RestMan_onQuitRequest.caller.name;
		};
		if(_RestMan.codeMonitoring) {
			_RestMan.promptSrvc.alert(null, "onQuitRequest caller:",
				callerName + "\n\n" + "Passed parameter: " + parameter);
		};
		if(aParam == "restoreUserSetCmd" || aParam == "handleExit"
			|| aParam == "handleCloseLastWin") {
			onQuitCmd(aParam);
			return;
		};
		function onQuitCmd(aParam) {
			if(_RestMan.codeMonitoring) {
				_RestMan.promptSrvc.alert(null, "onQuitCmd caller:",
					onQuitCmd.caller.name // arguments.callee.caller
					+ "\n\n" + "Passed parameter: " + parameter);
			};
			if(aParam == "restoreUserSetCmd") {
				restoreUserSettingsCmd();
				return;
			};
			function restoreUserSettingsCmd() {
				// We restore here user settings if necessary (if we have changed them).
				// We could do it easily at next startup, but there is a small issue: (Note: value = value of "browser.startup.page").
				// After our "restart no restore", there is a difference between "value = 1" and "value = 3": If the user
				// decide to "Restore Previous Session", in first case (value = 1) the session is restored as usual,
				// while in second case (value = 3) the initial tab is not overwritten but is merged with the
				// restored session.
				// This is because the condition where, "value = 3" and the option "Restore Previous Session" available,
				// is not included in pm's design.
				// With value = 0 there is no difference, but however, we use "value = 1" because this is pm's default.
				// Also, if "value = 3" (no dialogs on quit), if we set "value = 1", causes the quit or
				// warning dialog to be shown on quit. So, for an exact simulation of "value = 3", we also suppress temporarily
				// all pm's dialogs, by setting "warnOnQuit = false", if this is not already set by the user as a personal
				// preference, until next quit or "restart & restore".
				_RestMan.brPrefBranch.setIntPref("startup.page", 3);
				_RestMan.rmPrefBranch.setBoolPref("restore_UserSettings", false);
				if(_RestMan.restWarnOnQuit) {
					_RestMan.brPrefBranch.setBoolPref("warnOnQuit", true); // pm's default.
					_RestMan.rmPrefBranch.setBoolPref("restore_warnOnQuitPref", false);
				};
				if(_RestMan.codeMonitoring) {
					var setByUser = "";
					if(!_RestMan.restWarnOnQuit) {
						setByUser = " - (set by the user)";
					};
					_RestMan.promptSrvc.alert(null, "Restore startup/warnOnQuit settings:",
						"browser.startup.page: " + _RestMan.brPrefBranch.getIntPref("startup.page") +
						"\nbrowser.warnOnQuit: " + _RestMan.brPrefBranch.getBoolPref("warnOnQuit") +
						setByUser);
				};
			};
			if(_RestMan.restoreUserSet) {
				restoreUserSettingsCmd();
			};
			_RestMan.restmanSrvc.observerUnregister();
			if(_RestMan.appVer >=20 && _RestMan.brPrefBranch.getBoolPref("showQuitWarning")
				&& (aParam == "handleExit" || aParam == "handleCloseLastWin")) { // Somewhat arbitrary setting,
				// assuming the users which have enabled the quit dialog, most likely in most cases they want the
				// session to being restored automatically in next startup.
				_RestMan.brPrefBranch.setBoolPref("sessionstore.resume_session_once", true);
			};
			if(aParam == "handleExit") { // Passed by our js component on "File --> Exit".
				_RestMan.sessionSaveHelper(aParam);
			} else if(aParam == "handleCloseLastWin") { // On closing the last browser window which is private
				_RestMan.sessionSaveHelper();			// or all of its tabs are private.
			};
			return;
		};
		var loopCount = 1; // Normally, the dialog is detected instantly, but let's stay in the safe
						   // side, taking this precautionary measure for possible infinite loops..
		function onQuitDlg() {
			_RestMan.timer.initWithCallback(function() {
				if(_RestMan.dialogDetector()) {
					_RestMan.openedDialog.addEventListener("dialogaccept",
						function dlg_dialogaccept(){
							onQuitCmd();
							return;
						}, false);
					_RestMan.openedDialog.addEventListener("dialogextra1",
						function dlg_dialogextra1(){
							onQuitCmd();
							return;
						}, false);
					if(_RestMan.codeMonitoring) {
						_RestMan.promptSrvc.alert(null, "Dialog detection:",
						"Dialog: " + _RestMan.openedDialog + "\nDetected in  " + loopCount + "  attempts");
					};
					return;
				} else if(!_RestMan.openedDialog && loopCount <20) {
					loopCount++;
					onQuitDlg(); // arguments.callee
				} else if(!_RestMan.openedDialog && loopCount == 20) {
					// The dialog is not detected for some reason..
					onQuitCmd();
					return;
				};
			}, 15, ci.nsITimer.TYPE_ONE_SHOT);
		};
		if(callerName == "RestMan_winTabCounter") {
		// We are here on closing the last non-private browser window via close button
		// or Ctrl+Shift+w. If there was a dialog, it has been already accepted.
			onQuitCmd();
			return;
		};
		// We are here by our observer, on exiting pm via menu "File --> Exit",
		// on non-private browser mode, so maybe we have pm's quit dialog.
		// Trying to determine if we'll have the dialog:
		_RestMan.winTabCounter();
		if(_RestMan.allTabCount <2
			|| _RestMan.brPrefBranch.getIntPref("startup.page") == 3
			|| _RestMan.brPrefBranch.getBoolPref("sessionstore.resume_session_once")
			|| _RestMan.allWinPrivate
			|| _RestMan.allTabsPrivate
			|| !(_RestMan.brPrefBranch.getBoolPref("warnOnQuit"))
			|| !(_RestMan.brPrefBranch.getBoolPref("showQuitWarning"))
			) { // No pm's quit dialog. On exiting, if there are only private windows and
				// >=2 tabs, we show our warning dialog if this is selected (default)
				// as we do also in our restart process from private browsing mode.
				// Pm should also show a similar (optional) warning dialog, because
				// private windows can not be restored after closing them..
			if(_RestMan.codeMonitoring) {
				_RestMan.promptSrvc.alert(null, "No pm's quit dialog is expected", "All tabs: " + _RestMan.allTabCount +
										"\nVisible tabs: " + _RestMan.visTabCount);
			};
			onQuitCmd();
			return;
		} else { // We expect pm's quit dialog.
			if(_RestMan.codeMonitoring) {
				_RestMan.promptSrvc.alert(null, "Pm's quit dialog is expected", "All tabs: " + _RestMan.allTabCount +
										  "\nVisible tabs: " + _RestMan.visTabCount);
			};
			onQuitDlg();
		};
	},

	optionsDlgOnLoad: function() {
		_RestMan.sharedStrings();
		var prefWin = window.document.getElementById("restman-Options");
		var restmanBundle = _RestMan.strings.createBundle("chrome://restartmanager/locale/restartmanager.properties");
		var extensionsBundle = _RestMan.strings.createBundle("chrome://mozapps/locale/extensions/extensions.properties");
		var easternLocale = false;
		var openDlgChkbx = window.document.getElementById("openDlgChkbox");
		var radioTooltip = window.document.getElementById("radioTooltip");
		var iconCheckbox = window.document.getElementById("iconCheckbox");
		var shortcutVisState = window.document.getElementById("shortcutVis-state");
		var shortcutState = window.document.getElementById("shortcut-state");
		var invChkbx = window.document.getElementById("invCaches");
		var restmanVersion = window.document.getElementById("restmanVersion");
		var hChkbx = window.document.getElementById("highlight");
		var hChkbxLabel = window.document.getElementById("hiddenLabel2").value;
		openDlgChkbx.label = restmanBundle.GetStringFromName("openDlgChkbox-label");
		openDlgChkbx.setAttribute("accesskey", restmanBundle.GetStringFromName("openDlgChkbox-accesskey"));
		openDlgChkbx.setAttribute("tooltiptext", restmanBundle.GetStringFromName("openDlgChkbox-tooltip"));
		radioTooltip.setAttribute("tooltiptext", restmanBundle.GetStringFromName("iconIdea-tooltip"));
		shortcutVisState.setAttribute("tooltiptext", restmanBundle.GetStringFromName("shortcutIndication-tooltip"));
		invChkbx.setAttribute("tooltiptext", restmanBundle.GetStringFromName("invalidateCachesOnRestart-tooltip"));
		restmanVersion.setAttribute("value", extensionsBundle.formatStringFromName("aboutWindowVersionString", ["2.02"], 1));

		hChkbx.label = hChkbxLabel.replace(/ +\S+/,""); // Highlight All --> Highlight
		if(_RestMan.appVer >=3.5) {
			prefWin.setAttribute("title", _RestMan.optionsTitle);
		} else if(_RestMan.appVer <3.5) {
			if(prefWin) { // Only this method works in pm 1.5, 2 & 3.
				window.document.title = _RestMan.optionsTitle;
			};
		};
		window.document.getElementById("confirmCap").label = _RestMan.confirmLabel;
		var askCheckboxLabel = window.document.getElementById("askCheckbox");
		var neverAskStr = _RestMan.neverAskLabel; // We use this label as a reference, to check if the locale
												  // is an eastern one (e.g. Japanese, Chinese...) or not.
		if(_RestMan.appVer >=23) { // The label "Do not ask next time" in pm >=23 has access key with
			if((/(^|[^&])&([^&])/).test(neverAskStr) && // ampersand ("&"), recognized only by prompt service..
				!(/\(&.\)/).test(neverAskStr)) { // Should exists in eastern locales.
				askCheckboxLabel.label = RegExp.leftContext + RegExp.$1 + RegExp.$2 + RegExp.rightContext;
				askCheckboxLabel.setAttribute("accesskey", RegExp.$2);
				shortcutVisState.setAttribute("accesskey", shortcutVisState.label.match(/...$/).toString().charAt(0));
				shortcutState.setAttribute("accesskey", shortcutState.label.charAt(2));
				hChkbx.setAttribute("accesskey", hChkbxLabel.match(/^./)); // Highlight All --> H
			} else { // Probably it's an eastern locale...
				easternLocale = true;
				askCheckboxLabel.label = _RestMan.neverAskLabel.replace(/\(&.\)/,"");
				shortcutVisState.setAttribute("accesskey", "V");
				shortcutState.setAttribute("accesskey", "S");
				hChkbx.setAttribute("accesskey", "H");
			};
		} else if(_RestMan.appVer <23) {
			if((/. ./).test(neverAskStr)) { // If there are spaces, we assume it's not an eastern locale.
				askCheckboxLabel.label = _RestMan.neverAskLabel;
				askCheckboxLabel.setAttribute("accesskey", _RestMan.neverAskLabel.match(/^./));
				shortcutVisState.setAttribute("accesskey", shortcutVisState.label.match(/...$/).toString().charAt(0));
				shortcutState.setAttribute("accesskey", shortcutState.label.charAt(2));
				hChkbx.setAttribute("accesskey", hChkbxLabel.match(/^./)); // Highlight All --> H
			} else {
				easternLocale = true;
				askCheckboxLabel.label = _RestMan.neverAskLabel;
				shortcutVisState.setAttribute("accesskey", "V");
				shortcutState.setAttribute("accesskey", "S");
				hChkbx.setAttribute("accesskey", "H");
			};
		};
		askCheckboxLabel.parentNode.insertBefore(askCheckboxLabel, askCheckboxLabel.nextSibling);
		openDlgChkbx.parentNode.insertBefore(openDlgChkbx, openDlgChkbx.nextSibling);
		shortcutVisState.parentNode.insertBefore(shortcutVisState, shortcutVisState.nextSibling);
		shortcutState.parentNode.insertBefore(shortcutState, shortcutState.nextSibling);
		hChkbx.parentNode.insertBefore(hChkbx, hChkbx.nextSibling);
		if(easternLocale) {
			iconCheckbox.setAttribute("accesskey", "T");
			iconCheckbox.parentNode.insertBefore(iconCheckbox, iconCheckbox.nextSibling);
		};
		if(_RestMan.appVer >=20) {
			var rChkbx = window.document.getElementById("restoreCheckbox");
			var warn2Chkbx = window.document.getElementById("warnOnClosePrvChkbox");
			if(!easternLocale) {
				// Attempt to avoid the same access key with the first "Remove" label.
				rChkbx.setAttribute("accesskey", rChkbx.label.charAt(3));
				rChkbx.parentNode.insertBefore(rChkbx, rChkbx.nextSibling);
				warn2Chkbx.setAttribute("accesskey", window.document.getElementById("hiddenLabel1")
					.getAttribute("value").toLocaleLowerCase().match(/^./)); // "Close" --> "c"
			} else {
				warn2Chkbx.setAttribute("accesskey", "W");
			};
		};
		if(_RestMan.appVer >=6) {
			window.document.getElementById("btnStyle-label").value = _RestMan.appearanceLabel + ":";
		};
		if(_RestMan.appVer >=4) {
			Array.forEach(window.document.getElementsByClassName("restman-radioButton"),
				function(inEachBtn) {
					inEachBtn.setAttribute("iconStyle", _RestMan.iconStyle);
					if(_RestMan.invCachHghlght) {
						inEachBtn.setAttribute("highlightBtn", true);
					};
			});
			window.document.getElementById("restman-radioBtn2").setAttribute("iconStyle", _RestMan.iconStyle);
			if(_RestMan.invCachHghlght) {
				window.document.getElementById("restman-radioBtn2").setAttribute("highlightBtn", true);
			};
			invChkbx.addEventListener("command", function() {
				window.addEventListener("unload", function() {
					if(_RestMan.rmPrefBranch.getBoolPref("invalidate_caches")) {
						_RestMan.prefSrvc.getBranch("nglayout.").setBoolPref("debug.disable_xul_fastload", true);
						_RestMan.prefSrvc.getBranch("nglayout.").setBoolPref("debug.disable_xul_cache", true);
					} else {
						_RestMan.prefSrvc.getBranch("nglayout.").setBoolPref("debug.disable_xul_fastload", false);
						_RestMan.prefSrvc.getBranch("nglayout.").setBoolPref("debug.disable_xul_cache", false);
					};
				}, false);
			}, false);
		};
		var txtField = window.document.getElementById("shortcut-char");
		if(txtField.value == "" || txtField.value == " ") {
			window.setTimeout(function() {
				if(_RestMan.appVer >=13) {
					if(txtField) {
						txtField.reset(); // Restore default.
						txtField.dispatchEvent(new Event("input", {
							view: window, bubbles: true, isTrusted: true}));
						return true;
					};
				} else if(_RestMan.appVer <13) {
					var event = window.document.createEvent("UIEvents");
					event.initUIEvent("input", true, true, window, 1);
					if(txtField) {
						txtField.value = "r"; // Restore default.
						txtField.dispatchEvent(event);
						return true;
					};
				};
			}, 650);
		};
	},

	applyUISettings: function(aParam) {
		_RestMan.sharedStrings();
		var rmPrefObserver = {
			checkObsereverResponseDelay: null,
			checkObsereverResponse: function() {
			// Clicking already selected radio buttons, the pref observer
			// gets registered and stays there in standby.. Don't like this :-)
				checkObsereverResponseDelay = window.setTimeout(function() {
					if(_RestMan.dialogDetector() &&
						_RestMan.openedDialog.document.documentElement.getAttribute("windowtype") == "dialog:commonDialog") {
						return rmPrefObserver.checkObsereverResponse(); // If code monitoring is enabled, this reset/repeat
						// function is required in pm 1.5 & 2. Otherwise, on timeout expiration, the pref observer will be
						// removed if the first alert ("Registered") is not accepted within the specified timeout.
					};
					if(_RestMan.setPrefObserver) {
						rmPrefObserver.observerUnregister();
						_RestMan.setPrefObserver = false;
					};
				}, 500);
			},
			prefBranch2Observe: _RestMan.rmPrefBranch,
			observerRegister: function() {
				if(!_RestMan.setPrefObserver) {
					if(_RestMan.appVer <13) {
						this.prefBranch2Observe.QueryInterface(ci.nsIPrefBranch2); // Pm < 13.
					};
					this.prefBranch2Observe.addObserver("", this, false);
					_RestMan.setPrefObserver = true;
					if(_RestMan.codeMonitoring) {
						alert("applyUISettings pref observer:" + "\nRegistered");
					};
					this.checkObsereverResponse();
				};
			},
			observerUnregister: function() {
				if(_RestMan.setPrefObserver) {
					this.prefBranch2Observe.removeObserver("", this);
					_RestMan.setPrefObserver = false;
					if(_RestMan.codeMonitoring) {
						alert("applyUISettings pref observer:" + "\nUnregistered");
					};
				};
			},
			observe: function(aSubject, aTopic, aData) {
				if(_RestMan.codeMonitoring) {
					alert("Observe:" + "\nSubject: " + aSubject + "\nTopic: " + aTopic + "\nData: " + aData);
				};
				if(aTopic !== "nsPref:changed") {
					rmPrefObserver.observerUnregister();
					return false;
				};
				var winEnumerator = _RestMan.winMed.getEnumerator(null); // "null", because we want to be included
																		 // also our options dialog.
				while(winEnumerator.hasMoreElements()) {
					var winDoc = winEnumerator.getNext().document;
					var appMenu = winDoc.getElementById("appmenuPrimaryPane");
					var fileMenu = winDoc.getElementById("menu_FilePopup");
					var appMenuItem = winDoc.getElementById("restman-appMenuItem");
					var fileMenuItem = winDoc.getElementById("restman-fileMenuItem");
					var neighbor;
					var restartButton = winDoc.getElementById("restman-button");
					var restmanKeyset = winDoc.getElementById("restmanKeyset");
					var winDocType = winDoc.documentElement.getAttribute("windowtype");
					function rShortcutSetUp(aType, aParam, aValue) { // Constructor
						// Recreates the menu item as it was before its removal.
						var newMenuItemId;
						if(aType == "appMenuItem") {
							newMenuItemId = "restman-appMenuItem";
						} else if(aType == "fileMenuItem") {
							newMenuItemId = "restman-fileMenuItem";
						};
						var newMenuItem = winDoc.createElement("menuitem");
						newMenuItem.id = newMenuItemId;
						newMenuItem.setAttribute("class", "menuitem-iconic");
						newMenuItem.setAttribute("label", _RestMan.restartLabel);
						newMenuItem.setAttribute("accesskey", _RestMan.restartAccKey);
						newMenuItem.setAttribute("command", "cmd_restmanRestartApp");
						if(_RestMan.menuItemIcon) {
							newMenuItem.setAttribute("menuicon", true);
						};
						newMenuItem.setAttribute("iconStyle", _RestMan.iconStyle);
						if(_RestMan.invCachHghlght) {
							newMenuItem.setAttribute("highlightMenuItem", true);
						}; //
						var addRestartShortcut = _RestMan.rmPrefBranch.getBoolPref("restart_shortcut");
						var restartShortcutVis = _RestMan.rmPrefBranch.getBoolPref("restart_shortcutVis");
						if(addRestartShortcut) {
						// If true, recreate the shortcut as it was before its removal
						// or in accordance the new user settings.
							var shortcutKeyMod = _RestMan.shortcutKeyMod;
							var shortcutChar = _RestMan.shortcutChar;
							if(aParam == "Alt") {
								shortcutKeyMod = "alt";
							} else if(aParam == "Shift") {
								shortcutKeyMod = "shift";
							} else if(aParam == "Char") {
								shortcutChar = aValue;
							};
							var newRestmanKeyset = winDoc.createElement("keyset");
							newRestmanKeyset.id = "restmanKeyset";
							var newRestartShortcut = winDoc.createElement("key");
							newRestartShortcut.id = "key_restmanKeyCombo";
							newRestartShortcut.setAttribute("command", "cmd_restmanRestartApp");
							newRestartShortcut.setAttribute("modifiers", ("accel," + shortcutKeyMod));
							newRestartShortcut.setAttribute("key", shortcutChar);
							if(!winDoc.getElementById("restmanKeyset")) { // If we don't do this check,
							// the new "keyset" will be appended twice. This function is called twice,
							// once for app menu item and once for file menu item.
								winDoc.getElementById("main-window")
									.appendChild(newRestmanKeyset).appendChild(newRestartShortcut);
							};
							if(restartShortcutVis) {
								newMenuItem.setAttribute("key", "key_restmanKeyCombo");
							};
						}; //
						if(_RestMan.appVer >=22) {
							newMenuItem.setAttribute("appVer22-28", "");
						};
						return newMenuItem;
					};
					switch (aData) {
						case "openDlg_detection": // Enable/disable of opened dialogs detection.
							if(winDoc && winDocType == "navigator:browser") {
								var openDlgDetPref = this.prefBranch2Observe.getBoolPref("openDlg_detection");
								if(openDlgDetPref) {
									winDoc.defaultView._RestMan.detectOpenDlgs = true;
								} else {
									winDoc.defaultView._RestMan.detectOpenDlgs = false;
								}; // Update of all existing ChromeWindows.
							};
							break;
						case "menuItemIcon": // Add/remove the icon of our menu items.
							if(winDoc && winDocType == "navigator:browser") {
								var iconOnMenuItems = this.prefBranch2Observe.getBoolPref("menuItemIcon");
								if(!iconOnMenuItems) {
									if(appMenuItem) {
										appMenuItem.removeAttribute("menuicon");
									};
									if(fileMenuItem) {
										fileMenuItem.removeAttribute("menuicon");
									};
								} else {
									if(appMenuItem) {
										appMenuItem.setAttribute("menuicon", true);
									};
									if(fileMenuItem) {
										fileMenuItem.setAttribute("menuicon", true);
									};
								};
							};
							break;
						case "restart_shortcutVis": // Remove/restore the indication of our restart shortcut from restart menu items.
						case "restart_shortcut": // Enable/disable our restart shortcut.
							if(winDoc && winDocType == "navigator:browser") {
								if(restmanKeyset) {
									restmanKeyset.parentNode.removeChild(restmanKeyset);
								};
								if(appMenuItem) {
									neighbor = appMenuItem.nextSibling;
									appMenuItem.parentNode.removeChild(appMenuItem);
									appMenu.insertBefore(rShortcutSetUp("appMenuItem"), neighbor);
								};
								if(fileMenuItem) {
									neighbor = fileMenuItem.nextSibling;
									fileMenuItem.parentNode.removeChild(fileMenuItem);
									fileMenuItem = fileMenu.insertBefore(rShortcutSetUp("fileMenuItem"), neighbor); // Refreshes also the variable.
								};
							};
							break;
						case "shortcutKey_modifier": // Set key modifier for our restart shortcut (Alt/Option or Shift).
							if(winDoc && winDocType == "navigator:browser") {
								var keyModifier = "alt";
								var keyModifierPref = this.prefBranch2Observe.getCharPref("shortcutKey_modifier");
								if(keyModifierPref == "alt-option") {
									keyModifier = "Alt";
								} else if(keyModifierPref == "shift") {
									keyModifier = "Shift";
								};
								if(restmanKeyset) {
									restmanKeyset.parentNode.removeChild(restmanKeyset);
								};
								if(appMenuItem) {
									neighbor = appMenuItem.nextSibling;
									appMenuItem.parentNode.removeChild(appMenuItem);
									appMenu.insertBefore(rShortcutSetUp("appMenuItem", keyModifier), neighbor);
								};
								if(fileMenuItem) {
									neighbor = fileMenuItem.nextSibling;
									fileMenuItem.parentNode.removeChild(fileMenuItem);
									fileMenuItem = fileMenu.insertBefore(rShortcutSetUp("fileMenuItem", keyModifier), neighbor); // Refreshes also the variable.
								};
							};
							break;
						case "shortcut_char": // Set key letter for our restart shortcut.
							if(winDoc && winDocType == "navigator:browser") {
								var Char = this.prefBranch2Observe.getComplexValue("shortcut_char", ci.nsISupportsString).data;
								if(restmanKeyset) {
									restmanKeyset.parentNode.removeChild(restmanKeyset);
								};
								if(appMenuItem) {
									neighbor = appMenuItem.nextSibling;
									appMenuItem.parentNode.removeChild(appMenuItem);
									appMenu.insertBefore(rShortcutSetUp("appMenuItem", "Char", Char), neighbor);
								};
								if(fileMenuItem) {
									neighbor = fileMenuItem.nextSibling;
									fileMenuItem.parentNode.removeChild(fileMenuItem);
									fileMenuItem = fileMenu.insertBefore(rShortcutSetUp("fileMenuItem", "Char", Char), neighbor); // Refreshes also the variable.
								};
							};
							break;
						case "iconStyle": // Set icon style for our menu items.
							if(winDoc && (winDocType == "navigator:browser" || winDocType == "restman:options")) {
								var iconStylePref = this.prefBranch2Observe.getIntPref("iconStyle");
								if(restartButton) {
									restartButton.setAttribute("iconStyle", iconStylePref);
								};
								if(appMenuItem) {
									appMenuItem.setAttribute("iconStyle", iconStylePref);
								};
								if(fileMenuItem) {
									fileMenuItem.setAttribute("iconStyle", iconStylePref);
								};
								if(_RestMan.appVer >=4) {
									Array.forEach(winDoc.getElementsByClassName("restman-radioButton"),
										function(inEachBtn) {
											inEachBtn.setAttribute("iconStyle", iconStylePref);
									});
									var radioButton2 = winDoc.getElementById("restman-radioBtn2");
									if(radioButton2) {
										radioButton2.setAttribute("iconStyle", iconStylePref);
									};
								};
							};
							break;
						case "invalCaches_highlight": // Highlight/un-highlight all of our restart items.
							if(winDoc && (winDocType == "navigator:browser" || winDocType == "restman:options")) {
								var invalCachesHighlightPref = this.prefBranch2Observe.getBoolPref("invalCaches_highlight");
								if(invalCachesHighlightPref) {
									if(appMenuItem) {
										appMenuItem.setAttribute("highlightMenuItem", true);
									};
									if(fileMenuItem) {
										fileMenuItem.setAttribute("highlightMenuItem", true);
									};
									if(restartButton) {
										restartButton.setAttribute("highlightBtn", true);
									};
									if(_RestMan.appVer >=4) {
										Array.forEach(winDoc.getElementsByClassName("restman-radioButton"),
											function(inEachBtn) {
												inEachBtn.setAttribute("highlightBtn", true);
										});
										var radioButton2 = winDoc.getElementById("restman-radioBtn2");
										if(radioButton2) {
											radioButton2.setAttribute("highlightBtn", true);
										};
									};
								} else {
									if(appMenuItem) {
										appMenuItem.removeAttribute("highlightMenuItem");
									};
									if(fileMenuItem) {
										fileMenuItem.removeAttribute("highlightMenuItem");
									};
									if(restartButton) {
										restartButton.removeAttribute("highlightBtn");
									};
									if(_RestMan.appVer >=4) {
										Array.forEach(winDoc.getElementsByClassName("restman-radioButton"),
											function(inEachBtn) {
												inEachBtn.removeAttribute("highlightBtn");
										});
										var radioButton2 = winDoc.getElementById("restman-radioBtn2");
										if(radioButton2) {
											radioButton2.removeAttribute("highlightBtn");
										};
									};
								};
							};
							break;
						case "buttonStyle": // Set ours or default style for our toolbar button - appVer >= 4.
							if(winDoc && winDocType == "navigator:browser") {
								if(restartButton) {
									var btnStylePref = this.prefBranch2Observe.getIntPref("buttonStyle");
									if(btnStylePref == 1) {
										restartButton.setAttribute("class", "restman-toolbarbutton");
									} else if(btnStylePref == 0) {
										restartButton.setAttribute("class", "toolbarbutton-1");
									};
								};
							};
							break;
						case "allways_smallBtn": // Use small button allways - appVer >= 4.
							if(winDoc && winDocType == "navigator:browser") {
								if(restartButton) {
									if(this.prefBranch2Observe.getBoolPref("allways_smallBtn")) {
										restartButton.setAttribute("smallButton", true);
									} else {
										restartButton.removeAttribute("smallButton");
									};
								};
							};
							break;
					};
				};
				this.observerUnregister();
				window.clearTimeout(this.checkObsereverResponseDelay);
			}
		};
		if(aParam == "checkObserverState") {
			if(_RestMan.setPrefObserver) {
				if(_RestMan.codeMonitoring) {
					_RestMan.codeMonitoring = false;
					rmPrefObserver.observerUnregister();
				} else {
					rmPrefObserver.observerUnregister();
				};
			};
			return;
		} else {
			rmPrefObserver.observerRegister();
		};
	},

	animBtnIcon: function(aParam) {
		if(_RestMan.appVer >=4) {
			var targetBtn = window.document.getElementById("restman-button");
			if(targetBtn == undefined || targetBtn == null) {
				return false;
			};
			if(targetBtn.hasAttribute("rotateIcon")) { // Prevent retriggering of animation while it is running.
				return false;
			};
			var animDelay = window.setTimeout(function() {
								targetBtn.setAttribute("rotateIcon", true);
								animReset();
								targetBtn.removeEventListener("mouseout", animCancel, false);
							}, 550); // Time for "mouseover" on the button to trigger the animation.
			function animCancel() { // console.log("animCancel");
				window.clearTimeout(animDelay);
				targetBtn.removeEventListener("mouseout", animCancel, false);
			};
 			targetBtn.addEventListener("mouseout", animCancel, false);
			function animReset() { // We could listen for the "transitionend" event, but this is
			// dispatched also on button press, interrupting the animation. So, we use a timer instead..
				window.setTimeout(function() {
					targetBtn.removeAttribute("rotateIcon");
				}, 1550);
			};
   		};
	},

	launchOptionsDlg: function() {
		if(_RestMan.dialogDetector()) {
			_RestMan.openedDialog.focus();
			return;
		};
		window.openDialog("chrome://restartmanager/content/restmanOptions.xul","_blank",
			"chrome,titlebar,centerscreen,modal,dialog=yes,buttons=accept");
	}
};

/*
*********This is the end...,*******my only friend...,**********the end...***************************************
** This small project was a high school for me, to learn things :-)
** I got the idea for it when mozilla decided to remove the (very useful to me) restart dialog in Firefox 20..
****************************************************************************************************************
*/

